"""
The View classes render the data returned by a Source as a Panel
object.
"""
import sys

from io import BytesIO, StringIO
from weakref import WeakKeyDictionary

import numpy as np
import panel as pn
import param

from bokeh.models import NumeralTickFormatter
from panel.pane.base import PaneBase
from panel.pane.perspective import (
    THEMES as _PERSPECTIVE_THEMES, Plugin as _PerspectivePlugin,
)
from panel.param import Param
from panel.viewable import Viewable, Viewer

from ..base import Component, MultiTypeComponent
from ..config import _INDICATORS
from ..filters import Filter, ParamFilter
from ..panel import DownloadButton
from ..pipeline import Pipeline
from ..state import state
from ..transforms import SQLTransform, Transform
from ..util import catch_and_notify, is_ref, resolve_module_reference

DOWNLOAD_FORMATS = ['csv', 'xlsx', 'json', 'parquet']


class Download(Component, Viewer):

    color = param.Color(default='grey', allow_None=True, doc="""
      The color of the download button.""")

    hide = param.Boolean(default=False, doc="""
      Whether the download button hides when not in focus.""")

    format = param.ObjectSelector(default=None, objects=DOWNLOAD_FORMATS, doc="""
      The format to download the data in.""")

    kwargs = param.Dict(default={}, doc="""
      Keyword arguments passed to the serialization function, e.g.
      data.to_csv(file_obj, **kwargs).""")

    size = param.Integer(default=18, doc="""
      The size of the download button.""")

    view = param.Parameter(doc="Holds the current view.")

    # Specification configuration
    _internal_params = ['view']
    _required_keys = ['format']
    _validate_params = True

    @classmethod
    def validate(cls, spec, context=None):
        if isinstance(spec, str):
            spec = {'format': spec}
        return super().validate(spec, context)

    def __bool__(self):
        return self.format is not None

    def _table_data(self):
        if self.format in ('json', 'csv'):
            io = StringIO()
        else:
            io = BytesIO()
        data = self.view.get_data()
        if self.format == 'csv':
            data.to_csv(io, **self.kwargs)
        elif self.format == 'json':
            data.to_json(io, **self.kwargs)
        elif self.format == 'xlsx':
            data.to_excel(io, **self.kwargs)
        elif self.format == 'parquet':
            data.to_parquet(io, **self.kwargs)
        io.seek(0)
        return io

    def __panel__(self):
        filename = f'{self.view.pipeline.table}.{self.format}'
        return DownloadButton(
            callback=self._table_data, filename=filename, color=self.color,
            size=18, hide=self.hide
        )


class View(MultiTypeComponent, Viewer):
    """
    A View renders the data returned by a Source as a Viewable Panel
    object. The base class provides methods which query the Source for
    the latest data given the current filters and applies all
    specified `transforms`.

    Subclasses should use these methods to query the data and return
    a Viewable Panel object in the `get_panel` method.
    """

    controls = param.List(default=[], doc="""
        Parameters that should be exposed as widgets in the UI.""")

    download = param.ClassSelector(class_=Download, default=Download(), doc="""
        The download objects determines whether and how the source tables
        can be downloaded.""")

    pipeline = param.ClassSelector(class_=Pipeline, doc="""
        The data pipeline that drives the View.""")

    rerender = param.Event(default=False, doc="""
        An event that is triggered whenever the View requests a re-render.""")

    selection_group = param.String(default=None, doc="""
        Declares a selection group the plot is part of. This feature
        requires the separate HoloViews library.""")

    field = param.Selector(doc="The field being visualized.")

    view_type = None

    # Panel extension to load to render this View
    _extension = None

    # Parameters which reference fields in the table
    _field_params = ['field']

    _requires_source = True

    _selections = WeakKeyDictionary()

    _supports_selections = False

    __abstract = True

    def __init__(self, **params):
        self._cache = None
        self._ls = None
        self._panel = None
        self._updates = None
        refs = params.pop('refs', {})
        self.kwargs = {k: v for k, v in params.items() if k not in self.param}

        # Populate field selector parameters
        params = {k: v for k, v in params.items() if k in self.param}
        pipeline = params.pop('pipeline', None)
        if pipeline is None:
            raise ValueError("Views must declare a Pipeline.")
        fields = list(pipeline.schema)
        for fp in self._field_params:
            if isinstance(self.param[fp], param.Selector):
                self.param[fp].objects = fields
        pipeline.param.watch(self.update, 'data')
        super().__init__(pipeline=pipeline, refs=refs, **params)
        self.param.watch(self.update, [p for p in self.param if p not in ('rerender', 'selection_expr', 'name')])
        self.download.view = self
        if self.selection_group:
            self._init_link_selections()
        self.update()

    def __panel__(self):
        return pn.panel(pn.bind(lambda e: self.panel, self.param.rerender))

    def _init_link_selections(self):
        doc = pn.state.curdoc
        if self._ls is not None or doc is None:
            return
        if doc not in View._selections and self.selection_group:
            View._selections[doc] = {}
        self._ls = View._selections.get(doc, {}).get(self.selection_group)
        if self._ls is None:
            from holoviews.selection import link_selections
            self._ls = link_selections.instance()
            if self.selection_group:
                View._selections[doc][self.selection_group] = self._ls
        if 'selection_expr' in self.param:
            self._ls.param.watch(self._update_selection_expr, 'selection_expr')

    def _update_selection_expr(self, event):
        self.selection_expr = event.new

    @classmethod
    def _validate_filters(cls, *args, **kwargs):
        return cls._validate_list_subtypes('filters', Filter, *args, **kwargs)

    @classmethod
    def _validate_transforms(cls, *args, **kwargs):
        return cls._validate_list_subtypes('transforms', Transform, *args, **kwargs)

    @classmethod
    def _validate_sql_transforms(cls, *args, **kwargs):
        return cls._validate_list_subtypes('sql_transforms', SQLTransform, *args, **kwargs)

    @classmethod
    def _validate_pipeline(cls, *args, **kwargs):
        return cls._validate_str_or_spec('pipeline', Pipeline, *args, **kwargs)

    @classmethod
    def from_spec(cls, spec, source=None, filters=None, pipeline=None):
        """
        Resolves a View specification given the schema of the Source
        it will be filtering on.

        Parameters
        ----------
        spec: dict
            Specification declared as a dictionary of parameter values.
        pipeline: lumen.pipeline.Pipeline
            The Lumen pipeline driving this View. Must not be supplied
            if the spec contains a pipeline definition or reference.
        source: lumen.sources.Source
            The Source object containing the tables the View renders.
        filters: list(lumen.filters.Filter)
            A list of Filter objects which provide query values for
            the Source.

        Returns
        -------
        The resolved View object.
        """
        spec = spec.copy()
        resolved_spec, refs = {}, {}

        # Resolve pipeline
        if 'pipeline' in spec:
            if pipeline is not None:
                raise ValueError(
                    "Either specify the pipeline as part of the specification "
                    "or pass it in explicitly, not both."
                )
            pipeline = spec['pipeline']
            if isinstance(pipeline, str):
                resolved_spec['pipeline'] = state.pipelines[pipeline]
            else:
                resolved_spec['pipeline'] = Pipeline.from_spec(pipeline)
        else:
            overrides = {
                p: spec.pop(p) for p in Pipeline.param if p != 'name' and p in spec
            }
            for ts in ('transforms', 'sql_transforms'):
                if ts in overrides:
                    overrides[ts] = [Transform.from_spec(t) for t in overrides[ts]]
            if pipeline is None:
                pipeline = Pipeline(source=source, **overrides)
            elif overrides:
                pipeline = pipeline.chain(
                    filters=overrides.get('filters', []),
                    transforms=overrides.get('transforms', []),
                    sql_transforms=overrides.get('sql_transforms', [])
                )
            resolved_spec['pipeline'] = pipeline

        # Resolve View parameters
        view_type = View._get_type(spec.pop('type', None))
        for p, value in spec.items():
            if p in resolved_spec:
                continue
            elif p not in view_type.param:
                resolved_spec[p] = value
                continue
            parameter = view_type.param[p]
            if is_ref(value):
                refs[p] = value
                value = state.resolve_reference(value)
            if isinstance(parameter, param.ObjectSelector) and parameter.names:
                try:
                    value = parameter.names.get(value, value)
                except Exception:
                    pass
            resolved_spec[p] = value

        # Resolve download options
        download_spec = spec.pop('download', {})
        if isinstance(download_spec, str):
            download_spec = {'format': download_spec}
        resolved_spec['download'] = Download.from_spec(download_spec)

        view = view_type(refs=refs, **resolved_spec)

        # Resolve ParamFilter parameters
        for filt in (filters or []):
            if isinstance(filt, ParamFilter):
                if not isinstance(filt.parameter, str):
                    continue
                name, parameter = filt.parameter.split('.')
                if name == view.name and parameter in view.param:
                    filt.parameter = view.param[parameter]
        return view

    def __bool__(self):
        return self._cache is not None and len(self._cache) > 0

    @catch_and_notify
    def _update_panel(self, *events):
        """
        Updates the cached Panel object and returns a boolean value
        indicating whether a rerender is required.
        """
        if self._panel is not None:
            self._cleanup()
            updates = self._get_params()
            if updates is not None:
                self._panel.param.set_param(**updates)
                return False
        self._panel = self.get_panel()
        return True

    def _cleanup(self):
        """
        Method that is called on update.
        """

    def get_data(self):
        """
        Queries the Source for the specified table applying any
        filters and transformations specified on the View. Unlike
        `get_value` this should be used when multiple return values
        are expected.

        Returns
        -------
        DataFrame
            The queried table after filtering and transformations are
            applied.
        """
        if self._cache is not None:
            return self._cache
        if self.pipeline.data is None:
            self.pipeline._update_data()
        self._cache = data = self.pipeline.data
        return data

    def get_value(self, field=None):
        """
        Queries the Source for the data associated with a particular
        field applying any filters and transformations specified on
        the View. Unlike `get_data` this method returns a single
        scalar value associated with the field and should therefore
        only be used if only a single.

        Parameters
        ----------
        field: str (optional)
            The field from the table to return; if None uses
            field defined on the View.

        Returns
        -------
        object
            A single scalar value representing the current value of
            the queried field.
        """
        data = self.get_data()
        if not len(data) or field is not None and field not in data.columns:
            return None
        row = data.iloc[-1]
        return row[self.field if field is None else field]

    def get_panel(self):
        """
        Constructs and returns a Panel object which will represent a
        view of the queried table.

        Returns
        -------
        panel.Viewable
            A Panel Viewable object representing a current
            representation of the queried table.
        """
        return pn.panel(self.get_data())

    def update(self, *events, invalidate_cache=True):
        """
        Triggers an update in the View.

        Parameters
        ----------
        events: tuple
            param events that may trigger an update.
        invalidate_cache : bool
            Whether to clear the View's cache.
        """
        if invalidate_cache:
            self._cache = None
        stale = self._update_panel()
        if stale:
            self.param.trigger('rerender')

    def _get_params(self):
        return None

    @property
    def control_panel(self):
        return Param(
            self.param, parameters=self.controls, sizing_mode='stretch_width'
        )

    @property
    def panel(self):
        panel = self._panel
        if isinstance(panel, PaneBase):
            if len(panel.layout) == 1 and panel._unpack:
                panel = panel.layout[0]
            else:
                panel = panel._layout
        if self.download:
            return pn.Column(self.download, panel, sizing_mode='stretch_width')
        return panel

    @property
    def refs(self):
        refs = super().refs
        for c in self.controls:
            if c not in refs:
                refs.append(c)
        return refs


class Panel(View):
    """
    The `Panel` View is a very general purpose view that allows expressing
    arbitrary Panel objects as a specification. The Panel specification may
    be arbitrarily nested making it possible to specify entire layouts.
    Additionally the Panel specification also supports references,
    including standard source and variable references and a custom `$data`
    reference that inserts the current data of the `View`.

    :Example:

        type: panel
        spec:
            type: panel.layout.Column
            objects:
                - type: pn.pane.Markdown
                  object: '# My custom title'
                - type: pn.pane.DataFrame
                  object: $data
    """

    spec = param.Dict()

    view_type = 'panel'

    _requires_source = False

    def _resolve_spec(self, spec):
        if not isinstance(spec, dict) or 'type' not in spec:
            return spec
        spec = self.spec.copy()
        ptype = resolve_module_reference(spec.pop('type'), Viewable)
        params = {}
        for p, v in spec.items():
            if isinstance(v, dict) and 'type' in v:
                v = self._resolve_spec(v)
            elif isinstance(v, list):
                v = [self._resolve_spec(sv) for sv in v]
            elif is_ref(v):
                if v == '$data':
                    v = self.get_data()
                else:
                    v = state.resolve_reference(v)
            params[p] = v
        return ptype(**params)

    def get_panel(self):
        return self._resolve_spec(self.spec)


class StringView(View):
    """
    The StringView renders the latest value of the field as a HTML
    string with a specified fontsize.
    """

    font_size = param.String(default='24pt', doc="""
        The font size of the rendered field value.""")

    view_type = 'string'

    def get_panel(self):
        return pn.pane.HTML(**self._get_params())

    def _get_params(self):
        value = self.get_value()
        params = dict(self.kwargs)
        if value is None:
            params['object'] = 'No info'
        else:
            params['object'] = f'<p style="font-size: {self.font_size}">{value}</p>'
        return params


class IndicatorView(View):
    """
    The IndicatorView renders the latest field value as a Panel
    Indicator.
    """

    indicator = param.Selector(objects=_INDICATORS, doc="""
        The name of the panel Indicator type.""")

    label = param.String(doc="""
        A custom label to use for the Indicator.""")

    view_type = 'indicator'

    def __init__(self, **params):
        if 'indicator' in params and isinstance(params['indicator'], str):
            params['indicator'] = _INDICATORS[params['indicator']]
        super().__init__(**params)
        name = params.get('label', params.get('field', ''))
        self.kwargs['name'] = name

    def get_panel(self):
        return self.indicator(**self._get_params())

    def _get_params(self):
        params = dict(self.kwargs)
        if 'data' in self.indicator.param:
            params['data'] = self.get_data()
        else:
            value = self.get_value()
            if (not isinstance(value, (type(None), str)) and np.isnan(value)):
                value = None
            params['value'] = value
        return params


class hvPlotBaseView(View):

    kind = param.String(default=None, doc="The kind of plot, e.g. 'scatter' or 'line'.")

    x = param.Selector(doc="The column to render on the x-axis.")

    y = param.Selector(doc="The column to render on the y-axis.")

    by = param.ListSelector(doc="The column(s) to facet the plot by.")

    groupby = param.ListSelector(doc="The column(s) to group by.")

    __abstract = True

    def __init__(self, **params):
        import hvplot.pandas  # noqa
        if 'dask' in sys.modules:
            try:
                import hvplot.dask  # noqa
            except Exception:
                pass
        if 'by' in params and isinstance(params['by'], str):
            params['by'] = [params['by']]
        if 'groupby' in params and isinstance(params['groupby'], str):
            params['groupby'] = [params['groupby']]
        super().__init__(**params)


class hvPlotUIView(hvPlotBaseView):
    """
    The hvPlotUIView displays provides a component for exploring
    datasets using widgets.
    """

    view_type = 'hvplot_ui'

    def _get_args(self):
        from hvplot.ui import hvPlotExplorer
        params = {
            k: v for k, v in self.param.values().items()
            if k in hvPlotExplorer.param and v is not None and k != 'name'
        }
        return (self.get_data(),), dict(params, **self.kwargs)

    def __panel__(self):
        panel = self.get_panel()
        def ui(*events):
            panel._data = self.get_data()
            panel._plot()
            return panel
        return pn.bind(ui, self.param.rerender)

    def get_panel(self):
        from hvplot.ui import hvDataFrameExplorer
        args, kwargs = self._get_args()
        return hvDataFrameExplorer(*args, **kwargs)


class hvPlotView(hvPlotBaseView):
    """
    The hvPlotView renders the queried data as a bokeh plot generated
    with hvPlot. hvPlot allows for a concise declaration of a plot via
    its simple API.
    """

    opts = param.Dict(default={}, doc="HoloViews options to apply on the plot.")

    streaming = param.Boolean(default=False, doc="""
        Whether to stream new data to the plot or rerender the plot.""")

    selection_expr = param.Parameter(doc="""
        A selection expression caputirng the current selection applied
        on the plot.""")

    view_type = 'hvplot'

    _field_params = ['x', 'y', 'by', 'groupby']

    _ignore_kwargs = ['tables']

    _supports_selections = True

    def __init__(self, **params):
        self._stream = None
        self._linked_objs = []
        super().__init__(**params)

    def get_plot(self, df):
        processed = {}
        for k, v in self.kwargs.items():
            if k in self._ignore_kwargs:
                continue
            if k.endswith('formatter') and isinstance(v, str) and '%' not in v:
                v = NumeralTickFormatter(format=v)
            processed[k] = v
        if self.streaming:
            processed['stream'] = self._stream
        plot = df.hvplot(
            kind=self.kind, x=self.x, y=self.y, **processed
        )
        plot = plot.opts(**self.opts) if self.opts else plot
        if self.selection_group or 'selection_expr' in self._param_watchers:
            plot = self._link_plot(plot)
        return plot

    def _link_plot(self, plot):
        self._init_link_selections()
        if self._ls is None:
            return plot
        linked_objs = list(self._ls._plot_reset_streams)
        plot = self._ls(plot)
        self._linked_objs += [
            o for o in self._ls._plot_reset_streams if o not in linked_objs
        ]
        return plot

    def _cleanup(self):
        if self._ls is None:
            return
        for obj in self._linked_objs:
            reset = self._ls._plot_reset_streams.pop(obj)
            sel_expr = self._ls._selection_expr_streams.pop(obj)
            self._ls._cross_filter_stream.input_streams.remove(sel_expr)
            sel_expr.clear()
            sel_expr.source = None
            reset.clear()
            reset.source = None
        self._linked_objs = []

    def get_panel(self):
        return pn.pane.HoloViews(**self._get_params())

    def _get_params(self):
        df = self.get_data()
        if self.streaming:
            from holoviews.streams import Pipe
            self._stream = Pipe(data=df)
        return dict(object=self.get_plot(df))

    def update(self, *events, invalidate_cache=True):
        """
        Triggers an update in the View.

        Parameters
        ----------
        events: tuple
            param events that may trigger an update.
        invalidate_cache : bool
            Whether to clear the View's cache.
        """
        # Skip events triggered by a parameter change on this View
        own_parameters = [self.param[p] for p in self.param]
        own_events = events and all(
            isinstance(e.obj, ParamFilter) and
            (e.obj.parameter in own_parameters or
            e.new is self._ls.selection_expr)
            for e in events
        )
        if own_events:
            return
        if invalidate_cache:
            self._cache = None
        if not self.streaming or self._stream is None:
            stale = self._update_panel()
            if stale:
                self.param.trigger('rerender')
        else:
            self._stream.send(self.get_data())

class Table(View):
    """
    Renders a Source table using a Panel Table widget.
    """

    page_size = param.Integer(default=20, bounds=(1, None), doc="""
        Number of rows to render per page, if pagination is enabled.""")

    view_type = 'table'

    _extension = 'tabulator'

    def get_panel(self):
        return pn.widgets.tables.Tabulator(**self._get_params())

    def _get_params(self):
        return dict(value=self.get_data(), disabled=True, page_size=self.page_size,
                    **self.kwargs)


class DownloadView(View):
    """
    The Download View allows downloading the current table as a csv or
    xlsx file.
    """

    filename = param.String(default='data', doc="""
      Filename of the downloaded file.""")

    format = param.ObjectSelector(default=None, objects=DOWNLOAD_FORMATS, doc="""
      The format to download the data in.""")

    kwargs = param.Dict(default={}, doc="""
      Keyword arguments passed to the serialization function, e.g.
      data.to_csv(file_obj, **kwargs).""")

    view_type = 'download'

    _required_keys = ["format"]

    def __bool__(self):
        return True

    @catch_and_notify("Download failed")
    def _table_data(self):
        if self.format in ('json', 'csv'):
            io = StringIO()
        else:
            io = BytesIO()
        data = self.get_data()
        if self.format == 'csv':
            data.to_csv(io, **self.kwargs)
        elif self.format == 'json':
            data.to_json(io, **self.kwargs)
        elif self.format == 'xlsx':
            data.to_excel(io, **self.kwargs)
        elif self.format == 'parquet':
            data.to_parquet(io, **self.kwargs)
        io.seek(0)
        return io

    def get_panel(self):
        return pn.widgets.FileDownload(**self._get_params())

    def _get_params(self):
        filename = f'{self.filename}.{self.format}'
        return dict(filename=filename, callback=self._table_data, **self.kwargs)



class PerspectiveView(View):

    aggregates = param.Dict(None, allow_None=True, doc="""
        How to aggregate. For example {x: "distinct count"}""")

    columns = param.ListSelector(default=None, allow_None=True, doc="""
        A list of source columns to show as columns. For example ["x", "y"]""")

    computed_columns = param.ListSelector(default=None, allow_None=True, doc="""
        A list of computed columns. For example [""x"+"index""]""")

    column_pivots = param.ListSelector(None, allow_None=True, doc="""
        A list of source columns to pivot by. For example ["x", "y"]""")

    filters = param.List(default=None, allow_None=True, doc="""
        How to filter. For example [["x", "<", 3],["y", "contains", "abc"]]""")

    row_pivots = param.ListSelector(default=None, allow_None=True, doc="""
        A list of source columns to group by. For example ["x", "y"]""")

    selectable = param.Boolean(default=True, allow_None=True, doc="""
        Whether items are selectable.""")

    sort = param.List(default=None, doc="""
        How to sort. For example[["x","desc"]]""")

    plugin = param.ObjectSelector(default=_PerspectivePlugin.GRID.value, objects=_PerspectivePlugin.options(), doc="""
        The name of a plugin to display the data. For example hypergrid or d3_xy_scatter.""")

    theme = param.ObjectSelector(default='material', objects=_PERSPECTIVE_THEMES, doc="""
        The style of the PerspectiveViewer. For example material-dark""")

    view_type = 'perspective'

    _extension = 'perspective'

    _field_params = ['columns', 'computed_columns', 'column_pivots', 'row_pivots']

    def _get_params(self):
        df = self.get_data()
        param_values = dict(self.param.get_param_values())
        params = set(View.param) ^ set(PerspectiveView.param)
        kwargs = dict({p: param_values[p] for p in params}, **self.kwargs)
        return dict(object=df, toggle_config=False, **kwargs)

    def get_panel(self):
        return pn.pane.Perspective(**self._get_params())


class AltairView(View):

    chart = param.Dict(default={}, doc="Keyword argument for Chart.")

    x = param.Selector(doc="The column to render on the x-axis.")

    y = param.Selector(doc="The column to render on the y-axis.")

    marker = param.Selector(default='line', objects=[
        'area', 'bar', 'boxplot', 'circle', 'errorband', 'errorbar',
        'geoshape', 'image', 'line', 'point', 'rect', 'rule', 'square',
        'text', 'tick', 'trail'])

    encode = param.Dict(default={}, doc="Keyword arguments for encode.")

    mark = param.Dict(default={}, doc="Keyword arguments for mark.")

    transform = param.Dict(doc="""
        Keyword arguments for transforms, nested by the type of
        transform, e.g. {'bin': {'as_': 'binned', 'field': 'x'}}.""")

    project = param.Dict(doc="Keyword arguments for project.")

    properties = param.Dict(doc="Keyword arguments for properties.")

    view_type = 'altair'

    _extension = 'vega'

    def _transform_encoding(self, encoding, value):
        import altair as alt
        if isinstance(value, dict):
            value = dict(value)
            for kw, val in value.items():
                if kw == 'scale':
                    if isinstance(val, list):
                        val = alt.Scale(range=val)
                    else:
                        val = alt.Scale(**val)
                if kw == 'tooltip':
                    val = [alt.Tooltip(**v) for v in val]
                value[kw] = val
            value = getattr(alt, encoding.capitalize())(**value)
        return value

    def _get_params(self):
        import altair as alt
        df = self.get_data()
        chart = alt.Chart(df, **self.chart)
        mark = getattr(chart, f'mark_{self.marker}')(**self.mark)
        x = self._transform_encoding('x', self.x)
        y = self._transform_encoding('y', self.y)
        encode = {k: self._transform_encoding(k, v) for k, v in self.encode.items()}
        encoded = mark.encode(x=x, y=y, **encode)
        if self.transform:
            for key, kwargs in self.transform.items():
                encoded = getattr(encoded, f'transform_{key}')(**kwargs)
        if self.project:
            encoded = encoded.project(**self.project)
        if self.properties:
            encoded = encoded.properties(**self.properties)
        return dict(object=encoded, **self.kwargs)

    def get_panel(self):
        return pn.pane.Vega(**self._get_params())



import param

from panel.widgets import RangeSlider

from lumen.filters import Filter


def test_resolve_module_type():
    assert Filter._get_type('lumen.filters.base.Filter') is Filter


def test_widget_filter_link_unthrottled():
    wfilter = Filter.from_spec(
        {'type': 'widget', 'field': 'test', 'throttled': False},
        {'example': {
            'test': {
                'type': 'integer',
                'inclusiveMinimum': 0,
                'inclusiveMaximum': 2
            }
        }}
    )
    widget = wfilter.panel

    assert widget.value == (0, 2)

    widget.value = (1, 2)

    assert wfilter.value == (1, 2)

    wfilter.value = (2, 2)

    assert widget.value == (2, 2)

    widget.visible = False

    assert wfilter.visible == False

    widget.disabled = False

    assert widget.disabled == False


def test_widget_filter_link_throttled():
    wfilter = Filter.from_spec(
        {'type': 'widget', 'field': 'test'},
        {'example': {
            'test': {
                'type': 'integer',
                'inclusiveMinimum': 0,
                'inclusiveMaximum': 2
            }
        }}
    )
    widget = wfilter.panel

    assert widget.value == (0, 2)

    with param.edit_constant(widget):
        widget.value_throttled = (1, 2)

    assert wfilter.value == (1, 2)

    wfilter.value = (2, 2)

    assert widget.value == (2, 2)


def test_widget_filter_explicit_widget():
    wfilter = Filter.from_spec(
        {'type': 'widget', 'field': 'test', 'widget': 'panel.widgets.RangeSlider'},
        {'example': {
            'test': {
                'type': 'integer',
                'inclusiveMinimum': 1,
                'inclusiveMaximum': 3
            }
        }}
    )
    assert isinstance(wfilter.widget, RangeSlider)
    assert wfilter.widget.start == 1
    assert wfilter.widget.end == 3


def test_widget_filter_explicit_widget_to_spec():
    wfilter = Filter.from_spec(
        {'type': 'widget', 'field': 'test', 'widget': 'panel.widgets.RangeSlider'},
        {'example': {
            'test': {
                'type': 'integer',
                'inclusiveMinimum': 1,
                'inclusiveMaximum': 3
            }
        }}
    )
    assert wfilter.to_spec() == {'type': 'widget', 'field': 'test', 'widget': 'panel.widgets.slider.RangeSlider'}



"""
The Filter components supply query parameters used to filter the
tables returned by a Source.
"""
import types

import bokeh
import pandas as pd
import panel as pn
import param

from packaging.version import Version
from panel.util import classproperty

from ..base import MultiTypeComponent
from ..schema import JSONSchema
from ..state import state
from ..util import resolve_module_reference
from ..validation import ValidationError


class Filter(MultiTypeComponent):
    """
    A Filter provides a query which will be used to filter the data
    returned by a Source.
    """

    field = param.String(doc="The field being filtered.")

    label = param.String(doc="A label for the Filter.")

    schema = param.Dict(doc="""
        The JSON schema provided by the Source declaring information
        about the data to be filtered.""")

    shared = param.Boolean(default=False, doc="""
        Whether the filter is shared across all targets.""")

    sync_with_url = param.Boolean(default=True, doc="""
        Whether to sync the filter state with the URL parameters.""")

    table = param.String(default=None, doc="""
        The table being filtered. If None applies to all tables.""")

    value = param.Parameter(doc="The current filter value.")

    filter_type = None

    __abstract = True

    # Specification configuration
    _internal_params = ['name', 'schema']
    _requires_field = True

    def __init__(self, **params):
        super().__init__(**params)
        if state.app and state.app.config.sync_with_url and self.sync_with_url and pn.state.location:
            pn.state.location.sync(self, {'value': self.field}, on_error=self._url_sync_error)

    @classproperty
    def _required_keys(cls):
        return ['field'] if cls._requires_field else []

    def _url_sync_error(self, values):
        """
        Called when URL syncing errors.
        """

    ##################################################################
    # Public API
    ##################################################################

    @classmethod
    def from_spec(cls, spec, source_schema, source_filters=None):
        """
        Resolves a Filter specification given the schema of the Source
        (and optionally the table) it will be filtering on.

        Parameters
        ----------
        spec: dict or str
            Specification declared as a dictionary of parameter values.
        source_schema: dict
            A dictionary containing the JSON schema of the Source to
            be filtered on.
        source_filters: dict
            A dictionary of filters associated with the Source

        Returns
        -------
        The resolved Filter object.
        """
        if isinstance(spec, str):
            if source_filters is None:
                raise ValueError(f"Filter spec {repr(spec)} could not be resolved "
                                 "because source has not declared filters.")
            elif spec not in source_filters:
                raise ValueError(f"Filter spec {repr(spec)} could not be resolved, "
                                 f"available filters on the source include {list(source_filters)}.")
            return source_filters[spec]
        spec = dict(spec)
        filter_type = Filter._get_type(spec.pop('type', None))
        if not filter_type._requires_field:
            return filter_type(**spec)
        elif not 'field' in spec:
            raise ValueError('Filter specification must declare field to filter on.')
        field = spec['field']
        if 'label' not in spec:
            spec['label'] = field.title()
        table = spec.get('table')
        schema = None
        fields = []
        for key, table_schema in source_schema.items():
            if table is not None and key != table:
                continue
            schema = table_schema.get(field)
            fields.extend(list(table_schema))
            if schema is not None:
                break
        if not schema:
            raise ValueError(
                f"Source did not declare a schema for {field!r} filter. "
                f"Available fields include: {fields!r}."
            )
        return filter_type(schema={field: schema}, **spec)

    def to_spec(self, context=None):
        spec = super().to_spec(context=context)
        if spec['label'].lower() == spec['field']:
            del spec['label']
        return spec

    @property
    def panel(self):
        """
        Returns
        -------
        panel.Viewable or None
            A Panel Viewable object representing the filter.
        """
        return None

    @property
    def query(self):
        """
        Returns
        -------
        object
            The current filter query which will be used by the
            Source to filter the data.
        """

    @classmethod
    def validate(cls, spec, context=None):
        if isinstance(spec, str):
            return spec
        return super().validate(spec, context)


class ConstantFilter(Filter):
    """
    The ConstantFilter allows requesting a constant value from the
    Source.
    """

    filter_type = 'constant'

    @property
    def query(self):
        return self.value

    @property
    def panel(self):
        return None


class FacetFilter(Filter):
    """
    The FacetFilter allows faceting the data along some dimension to
    allow a single View to be exploded into multiple by grouping over
    one of the indexes.
    """

    filter_type = 'facet'

    @property
    def filters(self):
        field_schema = self.schema[self.field]
        if 'enum' in field_schema:
            values = field_schema['enum']
        elif field_schema['type'] == 'integer':
            values = list(range(field_schema['inclusiveMinimum'],
                                field_schema['inclusiveMaximum']+1))
        else:
            raise ValueError(f'Could not facet on field {field_schema!r}, '
                             'currently only enum and integer fields are '
                             'supported.')
        return [
            ConstantFilter(field=self.field, value=value, label=self.label)
            for value in values
        ]


class BaseWidgetFilter(Filter):

    default = param.Parameter(doc="""
        The default value to use on the widget.""")

    disabled = param.Boolean(default=False, doc="""
        Whether the filter should be disabled.""")

    throttled = param.Boolean(default=True, doc="""
       If the widget has a value_throttled parameter use that instead,
       ensuring that no intermediate events are generated, e.g. when
       dragging a slider.""")

    visible = param.Boolean(default=True, doc="""
        Whether the filter should be visible.""")

    __abstract__ = True

    def _url_sync_error(self, values):
        value = values['value']
        if value is self.widget.value:
            return
        elif isinstance(self.widget.param.value, param.Tuple):
            self.widget.value = tuple(value)
        else:
            raise ValueError(f'URL syncing failed, value {value!r} could not be applied.')

    @property
    def panel(self):
        widget = self.widget.clone()
        if self.throttled and 'value_throttled' in self.widget.param:
            widget.link(self.widget, value_throttled='value')
            self.widget.link(widget, value='value')
            self.widget.link(widget, visible='visible', disabled='disabled', bidirectional=True)
        else:
            self.widget.link(widget, value='value', visible='visible', disabled='disabled', bidirectional=True)
        return widget



class WidgetFilter(BaseWidgetFilter):
    """
    The WidgetFilter generates a Widget from the table schema provided
    by a Source and returns the current widget value to query the data
    returned by the Source.
    """

    empty_select = param.Boolean(default=True, doc="""
        Add an option to Select widgets to indicate no filtering.""")

    multi = param.Boolean(default=True, doc="""
        Whether to use a single-value or multi-value selection widget,
        e.g. for a numeric value this could be a regular slider or a
        range slider.""")

    widget = param.ClassSelector(class_=pn.widgets.Widget)

    filter_type = 'widget'

    _internal_params = ['name', 'schema']

    def __init__(self, **params):
        wtype = params.pop('widget', None)
        self._inferred_widget = wtype is None
        super().__init__(**params)
        self.widget = JSONSchema(
            schema=self.schema, sizing_mode='stretch_width', multi=self.multi,
            widgets={self.field: wtype} if wtype else {}
        )._widgets[self.field]
        if isinstance(self.widget, pn.widgets.Select) and self.empty_select:
            if self.widget.options[0] != ' ':
                self.widget.options.insert(0, ' ')
            self.widget.value = ' '
        self.widget.name = self.label
        self.widget.visible = self.visible
        self.widget.disabled = self.disabled
        self.widget.link(self, bidirectional=True, value='value', visible='visible', disabled='disabled')
        if self.default is not None:
            self.widget.value = self.default

    @classmethod
    def _validate_widget(cls, widget, spec, context):
        try:
            resolve_module_reference(widget, pn.widgets.Widget)
        except Exception:
            raise ValidationError(
                f'{cls.__name__} could not resolve widget module reference {widget!r}.', spec
            )
        else:
            return widget

    def to_spec(self, context=None):
        spec = super().to_spec(context=context)
        print(self._inferred_widget)
        if self._inferred_widget:
            del spec['widget']
        else:
            wtype = type(spec['widget'])
            spec['widget'] = f'{wtype.__module__}.{wtype.__name__}'
        return spec

    @property
    def query(self):
        if self.widget.value == ' ' and self.empty_select:
            return None
        if not hasattr(self.widget.param.value, 'serialize') or self.widget.value is None:
            return self.widget.value
        else:
            value = self.widget.param.value.serialize(self.widget.value)
            if isinstance(value, list) and isinstance(self.widget.value, tuple):
                return self.widget.value
            return value


class BinFilter(BaseWidgetFilter):
    """
    The BinFilter allows declaring a set of bins as a list of tuples
    and an optional set of labels of the same length.
    """

    bins = param.List(default=[], constant=True, doc="""
        A list of bins expressed as length two tuples.""")

    empty_select = param.Boolean(default=True, doc="""
        Add an option to Select widgets to indicate no filtering.""")

    labels = param.List(default=None, constant=True, doc="""
        A list of labels for each bin (optional).""")

    multi = param.Boolean(default=True, doc="""
        Whether to use a single-value or multi-value selection widget.""")

    filter_type = 'bins'

    def __init__(self, **params):
        super().__init__(**params)
        if self.multi:
            widget = pn.widgets.MultiSelect
        else:
            widget = pn.widgets.Select
        if self.labels:
            options = dict(zip(self.labels, [tuple(b) for b in self.bins]))
        else:
            options = {f'{l} - {u}': (l, u) for l, u in self.bins}
        options[' '] = None
        if self.default is None:
            value = [] if self.multi else None
        else:
            value = tuple(self.default)
        self.widget = widget(name=self.label, options=options, value=value)
        self.widget.link(self, value='value', visible='visible', disabled='disabled', bidirectional=True)

    @property
    def query(self):
        return self.widget.value


class BaseDateFilter(BaseWidgetFilter):

    mode = param.Selector(default='slider', objects=['slider', 'picker'], doc="""
        Whether to use a slider or a picker.""")

    multi = param.Boolean(default=True, doc="""
        Whether to use a single-value or multi-value/range selection widget.""")

    throttled = param.Boolean(default=True, doc="""
        Whether to throttle slider value changes.""")

    _as_date = False

    # Mapping from mode to a Panel widget, or to a function
    # that must return a tuple of a Panel widget and a dictionnary
    # of parameter values that will override the current ones.
    _widget_mode_mapping = {}

    __abstract = True

    def __init__(self, **params):
        super().__init__(**params)
        widget_type = self._widget_mode_mapping[self.mode]
        param_overrides = {}
        if isinstance(widget_type, types.FunctionType):
            widget_type, param_overrides = widget_type(self)
        if param_overrides:
            self.param.set_param(**param_overrides)
        self.widget = widget_type(**self._widget_kwargs(as_date=self._as_date))
        self.widget.link(self, value='value', visible='visible', disabled='disabled', bidirectional=True)

    def _widget_kwargs(self, as_date):
        field_schema = self.schema.get(self.field, {})
        kwargs = {
            'name': self.label,
            'start': pd.to_datetime(field_schema.get('inclusiveMinimum', None)),
            'end': pd.to_datetime(field_schema.get('inclusiveMaximum', None)),
        }
        if self.default is not None:
            kwargs['value'] = pd.to_datetime(self.default)
        if as_date:
            for key in ['value', 'start', 'end']:
                if key in kwargs and kwargs[key] is not None:
                    kwargs[key] = kwargs[key].date()
        return kwargs

    @property
    def panel(self):
        widget = self.widget.clone()
        if self.throttled and self.mode == 'slider':
            self.widget.link(widget, value='value')
            widget.link(self.widget, value_throttled='value')
        else:
            self.widget.link(widget, value='value', visible='visible', disabled='disabled', bidirectional=True)
        return widget

    @property
    def query(self):
        return self.widget.value


# DateFilter and DatetimeFilter both implement __new__ to return
# a Parameterized class that has the correct `value` Parameter type,
# which depends on the filter type (calendar date vs. datetime) and mode
# (slider vs picker).

class DateFilter(BaseDateFilter):
    """
    The DateFilter allows filtering by calendar dates, either by selecting
    a date or a date range, and with either a slider or a picker widget.
    """

    filter_type = 'date'

    def __new__(cls, **params):
        if params.get('multi', cls.param.multi.default):
            return _MultiCalendarDateFilter(**params)
        else:
            return _SingleCalendarDateFilter(**params)


class _SingleCalendarDateFilter(BaseDateFilter):

    _as_date = True

    _widget_mode_mapping = {
        'slider': pn.widgets.DateSlider,
        'picker': pn.widgets.DatePicker,
    }

    value = param.CalendarDate()


class _MultiCalendarDateFilter(BaseDateFilter):

    _as_date = True

    _widget_mode_mapping = {
        'slider': pn.widgets.DateRangeSlider,
        #'picker': pn.widgets.DateRangePicker,
    }

    value = param.CalendarDateRange()


class DatetimeFilter(BaseDateFilter):
    """
    The DatetimeFilter allows filtering by datetimes, either by selecting
    a datetime or a datetime range, and with either a slider or a picker widget.
    """

    filter_type = 'datetime'

    def __new__(cls, **params):
        if params.get('multi', cls.param.multi.default):
            return _MultiDatetimeFilter(**params)
        else:
            return _SingleDatetimeFilter(**params)


def _fallback_to_datetimepicker(inst):
    inst.param.warning(
        'Datetime multi/range slider filter not yet available, '
        'fallback to using a picker filter.'
    )
    return pn.widgets.DatetimePicker, {'mode': 'picker'}


class _SingleDatetimeFilter(BaseDateFilter):

    _as_date = False

    _widget_mode_mapping = {
        'slider': _fallback_to_datetimepicker,
        'picker': pn.widgets.DatetimePicker,
    }

    value = param.Date()


def _handle_datetimerangeslider(inst):
    if Version(bokeh.__version__) <= Version('2.4.3'):
        inst.param.warning(
            'Datetime multi/range slider filter requires Bokeh >= 2.4.3, '
            'fallback to using a picker filter.'
        )
        return pn.widgets.DatetimeRangePicker, {'mode': 'picker'}
    else:
        return pn.widgets.DatetimeRangeSlider, {}


class _MultiDatetimeFilter(BaseDateFilter):

    _as_date = False

    _widget_mode_mapping = {
        'slider': _handle_datetimerangeslider,
        'picker': pn.widgets.DatetimeRangePicker,
    }

    value = param.DateRange()


class ParamFilter(Filter):

    parameter = param.ClassSelector(default=None, class_=(param.Parameter, str), doc="""
        Reference to a Parameter on an existing View.""")

    filter_type = 'param'

    _requires_field = False

    def __init__(self, **params):
        super().__init__(**params)
        self.param.watch(self._attach_watcher, 'parameter')
        self._attach_watcher()

    def _update_value(self, event):
        self.value = event.new

    def _attach_watcher(self, *args):
        if isinstance(self.parameter, param.Parameter):
            self.parameter.owner.param.watch(self._update_value, self.parameter.name)



import os
import pathlib
import uuid

import panel as pn
import param

from panel.layout.base import ListLike
from panel.template.base import BasicTemplate

from lumen.config import config
from lumen.pipeline import Pipeline
from lumen.sources import Source
from lumen.state import state as lm_state

from .base import Wizard
from .config import ConfigEditor
from .dashboard import DashboardGallery
from .launcher import LauncherGallery
from .pipeline import PipelineGallery
from .sources import SourceGallery
from .state import state
from .targets import TargetEditor, TargetGallery, TargetGalleryItem
from .variables import VariablesEditor
from .views import ViewEditor, ViewGallery, ViewGalleryItem

CSS_RAW = """
i.fa.fa-plus:hover {
  color: var(--neutral-fill-color) !important;
}

.gallery-item:hover {
  box-shadow: 0 1px 5px;
}

.gallery-item {
  cursor: pointer;
}

.card-margin {
  height: calc(100% - 40px);
}

.bk-root {
  height: 100% !important;
}
"""

pn.extension('ace', 'perspective', 'tabulator', raw_css=[CSS_RAW])


class Builder(param.Parameterized):

    component_dir = param.Path(default=None)

    spec = param.Dict(default={})

    template = param.ClassSelector(class_=BasicTemplate)

    modal = param.ClassSelector(class_=ListLike)

    def __init__(self, **params):
        path = params['component_dir']
        dash_params, launcher_params, pipeline_params, source_params, target_params, var_params, view_params = (
            {}, {}, {}, {}, {}, {}, {}
        )
        dash_params['path'] = os.path.join(path, 'dashboards')
        launcher_params['path'] = os.path.join(path, 'launchers')
        pipeline_params['path'] = os.path.join(path, 'pipelines')
        source_params['path'] = os.path.join(path, 'sources')
        target_params['path'] = os.path.join(path, 'targets')
        var_params['path'] = os.path.join(path, 'variables')
        view_params['path'] = os.path.join(path, 'views')

        spec = params.pop('spec', {})
        if 'pipelines' not in spec:
            spec['pipelines'] = {}
        self.welcome = DashboardGallery(**dash_params)
        if 'launcher' in params:
            params['launcher'] = params['launcher'](spec=spec)
        super().__init__(spec=spec, **params)

        self.config = ConfigEditor(spec=self.spec['config'])
        self.variables = VariablesEditor(spec=self.spec['variables'], **var_params)
        lm_state.spec = state.spec = self.spec
        state.sources = self.sources = SourceGallery(spec=self.spec['sources'], **source_params)
        state.pipelines = self.pipelines = PipelineGallery(spec=self.spec['pipelines'], **pipeline_params)
        state.views = self.views = ViewGallery(**view_params)
        state.targets = self.targets = TargetGallery(spec=self.spec['targets'], **target_params)
        self.launcher = LauncherGallery(builder=self, **launcher_params)
        self.wizard = Wizard(items=[
            self.welcome, self.config, self.variables, self.sources,
            self.pipelines, self.views, self.targets, self.launcher
        ], sizing_mode='stretch_both')

        preview = pn.widgets.Button(name='Preview', width=100)
        preview.on_click(self._open_dialog)

        state.template.header.append(preview)
        if pn.state.user:
            logout = pn.widgets.Button(
                name='✖', width=40, css_classes=['logout'], margin=0,
                align='center'
            )
            logout.js_on_click(code="""
            window.location.href = '/logout'
            """)
            self.template.header.extend([
                pn.layout.HSpacer(),
                f'<b><font size="4.5em">User: {pn.state.user}</font></b>',
                logout
            ])
        self.template.main.append(self.wizard)
        self.template.modal.append(self.modal)

    def _open_dialog(self, event):
        self.wizard.preview.object = dict(state.spec)
        self.wizard.open_modal()

    @param.depends('welcome.spec', watch=True)
    def _update_spec(self):
        self.wizard.loading = True
        for key, item in self.welcome.spec.items():
            if isinstance(item, list):
                self.spec[key][:] = item
            else:
                if key in self.spec:
                    self.spec[key].clear()
                    self.spec[key].update(item)
                else:
                    self.spec[key] = item

        for name, source in self.sources.sources.items():
            config.root = str(pathlib.Path(__file__).parent)
            spec = source.spec
            if isinstance(source.spec, dict):
                spec = dict(source.spec)
                spec['name'] = name
                spec.pop('filters', None)

        for name, pipeline in self.pipelines.pipelines.items():
            config.root = str(pathlib.Path(__file__).parent)
            spec = pipeline.spec
            if isinstance(pipeline.spec, dict):
                spec = dict(pipeline.spec)
                spec['name'] = name
            try:
                lm_state.pipelines[name] = Pipeline.from_spec(spec)
            except Exception:
                pass

        self.config.param.trigger('spec')
        self.sources.param.trigger('spec')
        self.pipelines.param.trigger('spec')

        views, view_gallery = [], {}
        targets, target_items = [], {}
        for target in self.spec['targets']:
            if 'source' in target:
                source_spec = target['source']
                if isinstance(source_spec, dict):
                    source_spec = dict(source_spec, cache_dir=None)
                    source_spec.pop('filters', None)
                source = Source.from_spec(source_spec)
            elif 'pipeline' in target:
                pipeline_spec = target['pipeline']
                pipeline = Pipeline.from_spec(pipeline_spec)
            view_specs = target['views']
            if isinstance(view_specs, list):
                view_specs = [(f"{view['type']}: {view['table']}", view) for view in view_specs]
            else:
                view_specs = view_specs.items()
            target_views = []
            for name, view in view_specs:
                view = dict(view)
                view_type = view.pop('type')
                view_editor = ViewEditor(
                    view_type=view_type, name=name, source_obj=source,
                    spec=view, pipeline_obj=pipeline
                )
                view_gallery[name] = ViewGalleryItem(
                    editor=view_editor, name=name, selected=True
                )
                views.append(view_editor)
                target_views.append(name)

            target_editor = TargetEditor(spec=target, views=target_views)
            item = TargetGalleryItem(spec=target, editor=target_editor)
            targets.append(target_editor)
            target_items[f'target{uuid.uuid4().hex}'] = item

        self.targets.param.set_param(targets=targets, items=target_items)
        self.views.param.set_param(views=views, items=view_gallery)
        self.welcome.ready = True
        self.wizard.loading = False

    def servable(self, title='Lumen Builder'):
        self.template.servable(title=title)



import os
import tempfile

from unittest.mock import Mock

import pandas as pd
import panel as pn
import pytest

from bokeh.document import Document

from lumen.config import config
from lumen.sources import FileSource, Source
from lumen.state import state
from lumen.variables import Variables


@pytest.fixture
def set_root():
    root = config._root
    def _set_root(root):
        config._root = root
    yield _set_root
    config._root = root

@pytest.fixture
def make_filesource():
    root = config._root
    def create(root, **kwargs):
        config._root = root
        source = FileSource(tables={'test': 'test.csv'},  kwargs={'parse_dates': ['D']}, **kwargs)
        state.sources['original'] = source
        return source
    yield create
    config._root = root
    for source in state.global_sources.values():
        source.clear_cache()
    state.global_sources.clear()

@pytest.fixture
def make_variable_filesource():
    root = config._root
    def create(root, **kwargs):
        config._root = root
        state._variables[None] = Variables.from_spec({'tables': {'type': 'constant', 'default': {'test': 'test.csv'}}})
        source = Source.from_spec(dict({
            'type': 'file',
            'tables': '$variables.tables',
            'kwargs': {'parse_dates': ['D']}
        }))
        state.sources['original'] = source
        return source
    yield create
    config._root = root
    for source in state.global_sources.values():
        source.clear_cache()
    state.global_sources.clear()
    state._variables.clear()

@pytest.fixture
def mixed_df():
    yield pd._testing.makeMixedDataFrame()

@pytest.fixture
def yaml_file():
    tf = tempfile.NamedTemporaryFile(mode='w', suffix='.yaml', delete=False)
    yield tf
    tf.close()
    os.unlink(tf.name)

@pytest.fixture
def state_userinfo():
    mock_state = Mock()
    mock_state.user_info = {'email': 'lumen@holoviz.org', 'user': 'lumen'}
    mock_state.user = 'lumen'
    state = pn.state
    pn.state = mock_state
    yield
    pn.state = state

@pytest.fixture(autouse=True)
def clear_state():
    "Clear the state after the execution of each test"
    yield
    state.global_sources.clear()
    state.global_filters.clear()
    state._spec.clear()
    state._loading.clear()
    state._sources.clear()
    state._filters.clear()
    state._variables.clear()

@pytest.fixture
def document():
    doc = Document()
    with pn.io.server.set_curdoc(doc):
        yield

@pytest.fixture
def cachedir():
    tmp_dir = tempfile.TemporaryDirectory()
    yield tmp_dir.name
    tmp_dir.cleanup()


@pytest.fixture
def expected_filtered_df(source_tables, table_column_value_type):
    table, column, value, type = table_column_value_type
    df = source_tables[table]

    if type == 'single_value':
        if value is None:
            df = df[df[column].isnull()]
        else:
            df = df[df[column] == value]

    elif type == 'range':
        begin, end = value
        df = df[(df[column] >= begin) & (df[column] <= end)]

    elif type == 'range_list':
        conditions = False
        for range in value:
            begin, end = range
            conditions |= ((df[column] >= begin) & (df[column] <= end))
        df = df[conditions]

    elif type == 'list':
        df = df[df[column].isin(value)]

    elif type == 'date':
        df = df[df[column] == pd.to_datetime(value)]

    elif type == 'date_range':
        begin, end = value
        df = df[(df[column] >= pd.to_datetime(begin)) & (df[column] <= pd.to_datetime(end))]

    return df

@pytest.fixture
def penguins_file(tmp_path):
    # created with
    # df = pd.read_csv(url).sample(5).reset_index(drop=True).to_csv()
    fn = tmp_path / "penguins.csv"
    penguins = """
    ,species,island,bill_length_mm,bill_depth_mm,flipper_length_mm,body_mass_g,sex,year
    0,Gentoo,Biscoe,46.7,15.3,219.0,5200.0,male,2007
    1,Adelie,Dream,42.2,18.5,180.0,3550.0,female,2007
    2,Gentoo,Biscoe,51.1,16.3,220.0,6000.0,male,2008
    3,Adelie,Biscoe,45.6,20.3,191.0,4600.0,male,2009
    4,Adelie,Torgersen,37.8,17.1,186.0,3300.0,,2007
    """
    fn.write_text(penguins)
    return fn



import os
import pathlib

import panel as pn
import param
import requests
import yaml

from panel.reactive import ReactiveHTML

from lumen.sources import Source
from lumen.state import state as lm_state
from lumen.util import catch_and_notify

from .base import WizardItem
from .fast import FastComponent
from .gallery import Gallery, GalleryItem
from .state import state

ASSETS_DIR = pathlib.Path(__file__).parent / 'assets'


class SourceEditor(FastComponent):

    cache_dir = param.String(label="Cache directory (optional)", precedence=1, doc="""
        Enter a relative cache directory.""")

    form = param.Parameter()

    preview = param.Parameter()

    source_type = param.String(default="")

    shared = param.Boolean(default=False, precedence=1, doc="""
        Whether the Source can be shared.""")

    sizing_mode = param.String(default='stretch_width')

    source = param.Parameter(precedence=-1)

    spec = param.Dict(precedence=-1)

    resize = param.Integer(default=0)

    margin = param.Integer(default=10)

    _preview_display = param.String(default='none')

    _scripts = {
        'resize': ['window.dispatchEvent(new Event("resize"))']
    }

    _template = """
    <span style="font-size: 2em"><b>{{ name }} - {{ source_type }}</b></span>
    <div id="preview-area" style="display: ${_preview_display}; margin-top: 4.5em;">
      <form id="form" role="form" style="flex: 30%; max-width: 250px; line-height: 2em; margin-top">${form}</form>
      <div id="preview" style="flex: auto; margin-left: 1em; overflow-y: auto;">${preview}</div>
    </div>
    """

    _default_thumbnail = ASSETS_DIR / 'source.png'

    def __new__(cls, **params):
        if cls is not SourceEditor:
            return super().__new__(cls)
        editors = param.concrete_descendents(cls)
        source_type = params.get('spec', {}).get('type')
        for editor in editors.values():
            if editor.source_type == source_type:
                return super().__new__(editor)
        return super().__new__(cls)

    def __init__(self, **params):
        spec = params.pop('spec', {})
        params.update(**{
            k: v for k, v in spec.items() if k in self.param and k not in params
        })
        self._thumbnail = params.pop('thumbnail', None)
        super().__init__(spec=spec, **params)
        self.form = pn.Column(sizing_mode='stretch_width')
        theme = 'midnight' if getattr(pn.config, 'theme', 'default') == 'dark' else 'simple'
        self.preview = pn.widgets.Tabulator(
            sizing_mode='stretch_width', pagination='remote', page_size=12,
            theme=theme, height=400
        )
        self._select_table = pn.widgets.Select(
            name='Select table', margin=0, sizing_mode='stretch_width'
        )
        self._load_table = pn.widgets.Button(
            name='Load table', sizing_mode='stretch_width', margin=(15, 0)
        )
        def load_table(event):
            self.preview.value = self._source.get(self._select_table.value)
        self._load_table.on_click(load_table)
        self.form[:] = [self._select_table, self._load_table]

    @property
    def thumbnail(self):
        if self._thumbnail:
            return self._thumbnail
        return self._default_thumbnail

    def _update_spec(self, *events):
        for event in events:
            self.spec[event.name] = event.new

    def _preview(self, event):
        self._preview_display = 'flex' if self._preview_display == 'none' else 'none'
        self._source = Source.from_spec(self.spec)
        self._select_table.options = self._source.get_tables()
        self.resize += 1

    def _save(self):
        pass


class SourceGalleryItem(GalleryItem):

    editor = param.ClassSelector(class_=SourceEditor, precedence=-1)

    thumbnail = param.Filename()

    def __init__(self, **params):
        if 'description' not in params:
            params['description'] = ''
        super().__init__(**params)
        self.view = pn.pane.PNG(self.thumbnail, height=200, max_width=300, align='center')
        self._modal_content = [self.editor]

    @param.depends('selected', watch=True)
    def _add_spec(self):
        sources = state.spec['sources']
        if self.selected:
            spec = self.spec.copy()
            spec.pop('metadata', None)
            source = Source.from_spec(dict(spec, name=self.name))
            lm_state.sources[self.name] = source
            sources[self.name] = spec
        elif self.name in sources:
            del sources[self.name]
            del lm_state.sources[self.name]


class SourceGallery(WizardItem, Gallery):
    """
    Select the data sources to add to your dashboard specification.
    """

    path = param.Foldername()

    sources = param.Dict(default={}, doc="The list of sources added to the dashboard.", precedence=-1)

    _template = """
    <span style="font-size: 1.5em">Sources</span>
    <fast-divider></fast-divider>
    <span style="font-size: 1.2em; font-weight: bold;">{{ __doc__ }}</p>
    <div id="items" style="margin: 1em 0; display: flex; flex-wrap: wrap; gap: 1em;">
    {% for item in items.values() %}
      <fast-card id="source-container" style="width: 350px; height: 400px;">
        ${item}
      </fast-card>
    {% endfor %}
      <fast-card id="sources-container-new" style="height: 400px; width: 350px; padding: 1em;">
        <div style="display: grid;">
          <span style="font-size: 1.25em; font-weight: bold;">Add new source</span>
          <i id="add-button" onclick="${_open_modal}" class="fa fa-plus" style="font-size: 14em; margin: 0.2em auto;" aria-hidden="true"></i>
        </div>
      </fast-card>
    </div>
    """

    _gallery_item = SourceGalleryItem

    _editor_type = SourceEditor

    def __init__(self, **params):
        super().__init__(**params)
        for name, item in self.items.items():
            self.sources[name] = item.editor
        self._editor = SourcesEditor(spec={}, margin=10)
        self._save_button = pn.widgets.Button(name='Save sources')
        self._save_button.on_click(self._save_sources)
        self._modal_content = [self._editor, self._save_button]

    @param.depends('spec', watch=True)
    def _update_params(self):
        for name, source in self.spec.items():
            self.sources[name] = editor = SourceEditor(name=name, spec=source)
            self.items[name] = SourceGalleryItem(
                name=name, spec=source, selected=True, editor=editor,
                thumbnail=editor.thumbnail
            )
        self.param.trigger('sources')
        self.param.trigger('items')

    def _add_source(self, event):
        state.modal[:] = self._modal_content
        state.template.open_modal()

    def _save_sources(self, event):
        for name, source in self._editor.sources.items():
            path = pathlib.Path(self.path) / f'{name}.yaml'
            with open(path, 'w', encoding='utf-8') as f:
                f.write(yaml.dump(source.spec))
            self.spec[name] = source.spec
            item = SourceGalleryItem(
                name=name, spec=source.spec, margin=0, selected=True,
                editor=source, thumbnail=source.thumbnail
            )
            lm_state.sources[name] = Source.from_spec(dict(source.spec, name=name))
            self.items[name] = item
            self.sources[name] = source
        self.param.trigger('items')
        self.param.trigger('sources')
        self._editor.sources = {}
        state.template.close_modal()



class IntakeSourceEditor(SourceEditor):
    """
    Declare the Intake catalog either by entering the URI of the catalog file, uploading a file or manually entering the catalog file.
    """

    uri = param.String(precedence=1)

    source_type = param.String(default='intake', readonly=True)

    editor = param.Parameter(precedence=-1)

    upload = param.Parameter(precedence=-1)

    _template = """
    <span style="font-size: 1.5em">{{ name }} - Intake Source</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <form>
      <div style="display: flex; flex-wrap: wrap;">
        <div style="flex: 25%; min-width: 200px; margin-right: 1em;">
          <div style="display: grid;">
            <label for="URI"><b>URI</b></label>
            <fast-text-field id="uri" placeholder="Enter a URI" value="${uri}">
            </fast-text-field>
          </div>
          <div id="upload" style="display: grid; margin-top: 1em;">${upload}</div>
        </div>
        <div style="flex: 70%; min-width: 600px; display: block">
          <label for="catalog"><b>Catalog</b></label>
          <div id="catalog">${editor}</div>
        </div>
      </div>
      <div style="display: flex; margin-top: 1em;">
        <div style="display: grid; margin-right: 1em;">
          <label for="cache_dir"><b>{{ param.cache_dir.label }}</b></label>
          <fast-text-field id="cache_dir" placeholder="{{ param.cache_dir.doc }}" value="${cache_dir}" style="min-width: 300px;">
          </fast-text-field>
        </div>
        <div style="display: grid;">
          <label for="shared"><b>{{ param.shared.label }}</b></label>
          <fast-checkbox id="shared" value="${shared}"></fast-checkbox>
        </div>
      </div>
    </form>
    <fast-button id="preview-button" onclick="${_preview}" style="position: absolute; right: 5px; margin-top: 1.5em; z-index: 100;">
      Preview
    </fast-button>
    <fast-divider></fast-divider>
    <div id="preview-area" style="display: ${_preview_display}; margin-top: 4.5em;">
      <form id="form" role="form" style="flex: 30%; max-width: 250px; line-height: 2em;">${form}</form>
      <div id="preview" style="flex: auto; margin-left: 1em; overflow-y: auto;">${preview}</div>
    </div>
    """

    _dom_events = {'cache_dir': ['keyup'], 'uri': ['keyup']}

    _default_thumbnail = pathlib.Path(__file__).parent / 'assets' / 'intake.png'

    def __init__(self, **params):
        import lumen.sources.intake  # noqa
        params.pop('source_type', None)
        self.editor = pn.widgets.Ace(language='yaml', theme='dracula', margin=0, sizing_mode='stretch_width')
        self.upload = pn.widgets.FileInput(sizing_mode='stretch_width', margin=0)
        super().__init__(**params)

    @param.depends('upload.value', watch=True)
    def _upload_catalog(self):
        self.editor.value = self.upload.value.decode('utf-8')

    @param.depends('editor.value', watch=True)
    def _update_catalog(self):
        self.spec['catalog'] = yaml.safe_load(self.editor.value)

    @catch_and_notify
    @param.depends('uri', watch=True)
    def _load_file(self):
        uri = os.path.expanduser(self.uri)
        if os.path.isfile(uri):
            with open(uri) as f:
                self.editor.value = f.read()
        else:
            self.editor.value = requests.get(self.uri).content


class IntakeDremioSourceEditor(SourceEditor):
    """
    Provide a Dremio URI.
    """

    cert = param.String(default=None)

    load_schema = param.Boolean(default=False)

    tls = param.Boolean(doc="Enable TLS")

    uri = param.String(doc="Enter a URI")

    username = param.String(doc="Enter a username")

    password = param.String(doc="Enter a password")

    source_type = param.String(default='intake_dremio', readonly=True)

    _template = """
    <span style="font-size: 1.5em">{{ name }} - Intake Dremio Source</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <form>
      <div style="display: flex; flex-wrap: wrap;">
        <div style="flex: 25%; min-width: 200px; margin-right: 1em;">
          <div style="display: grid;">
            <label for="URI"><b>URI</b></label>
            <fast-text-field id="uri" placeholder="Enter a URI" value="${uri}">
            </fast-text-field>
          </div>
          <div style="display: grid;">
            <label for="username"><b>Username</b></label>
            <fast-text-field id="username" placeholder="Enter your username" value="${username}">
            </fast-text-field>
          </div>
          <div style="display: grid;">
            <label for="password"><b>Password</b></label>
            <fast-text-field id="password" type="password" placeholder="Enter password" value="${password}">
            </fast-text-field>
          </div>
          <fast-checkbox id="load_schema" checked="${load_schema}">Load schema</fast-checkbox>
          <fast-checkbox id="tls" checked="${tls}">Enable TLS</fast-checkbox>
          <div style="display: grid;">
            <label for="cert"><b>Certificate</b></label>
            <fast-text-field id="cert" disabled=${tls} placeholder="Enter path to a certificate" value="${cert}">
            </fast-text-field>
          </div>
        </div>
      </div>
      <div style="display: flex; margin-top: 1em;">
        <div style="display: grid; margin-right: 1em;">
          <label for="cache_dir"><b>{{ param.cache_dir.label }}</b></label>
          <fast-text-field id="cache_dir" placeholder="{{ param.cache_dir.doc }}" value="${cache_dir}" style="min-width: 300px;">
          </fast-text-field>
        </div>
        <div style="display: grid;">
          <label for="shared"><b>{{ param.shared.label }}</b></label>
          <fast-checkbox id="shared" value="${shared}"></fast-checkbox>
        </div>
      </div>
    </form>
    <fast-button id="preview-button" onclick="${_preview}" style="position: absolute; right: 5px; margin-top: 1.5em; z-index: 100;">
      Preview
    </fast-button>
    <fast-divider></fast-divider>
    <div id="preview-area" style="display: ${_preview_display}; margin-top: 4.5em;">
      <form id="form" role="form" style="flex: 30%; max-width: 250px; line-height: 2em;">${form}</form>
      <div id="preview" style="flex: auto; margin-left: 1em; overflow-y: auto;">${preview}</div>
    </div>
    """

    _dom_events = {'uri': ['keyup'], 'username': ['keyup'], 'password': ['keyup']}

    _default_thumbnail = pathlib.Path(__file__).parent / 'assets' / 'intake.png'

    def __init__(self, **params):
        import lumen.sources.intake  # noqa
        super().__init__(**params)

    @param.depends('cert', 'load_schema', 'tls', 'uri', 'password', 'username', watch=True)
    def _update_spec(self):
        for p in ('cert', 'load_schema', 'tls', 'uri', 'password', 'username'):
            self.spec[p] = getattr(self, p)



class FileSourceTable(ReactiveHTML):

    uri = param.String(doc="Enter a URI")

    margin = param.Integer(default=0)

    remove = param.Boolean(default=False, precedence=-1)

    sizing_mode = param.String(default='stretch_width')

    _template = """
    <form style="display: flex; flex-wrap: wrap; margin-top: 0.5em;">
      <div style="flex: 25%; min-width: 150px; display: grid; margin-right: 1em;">
        <label for="name"><b>Table Name</b></label>
        <fast-text-field id="name" placeholder="Enter a name" value="${name}"></fast-text-field>
      </div>
      <div style="flex: 60%; min-width: 300px; display: grid; margin-right: 1em;">
        <label for="uri"><b>URI</b></label>
        <fast-text-field id="uri" placeholder="{{ param.uri.doc }}" value="${uri}"></fast-text-field>
      </div>
      <fast-button style="width: 20px; margin-top: auto;" id="remove-source" onclick="${_remove}" appearance="accent">
        <b>－</b>
      </fast-button>
    </form>
    """

    _dom_events = {'uri': ['keyup'], 'name': ['keyup']}

    def _remove(self, event):
        self.remove = True


class FileSourceEditor(SourceEditor):
    """
    Declare a list of tables by providing a name and a URI to a local or remote file for each.
    """

    source_type = param.String(default='file', readonly=True)

    table_editors = param.List()

    kwargs = param.Dict(default={})

    _template = """
    <span style="font-size: 1.5em">{{ name }} - File Source</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <div id="tables">
      <fast-button id="add-table" onclick="${_add_table}" appearance="outline" style="float: right">
        <b>+</b>
      </fast-button>
      <span style="font-size: 1.2em; margin: 1em 0;"><b>Tables</b></span>
      ${table_editors}
    </div>
    <div style="display: flex; margin-top: 1em;">
      <div style="display: grid; margin-right: 1em;">
        <label for="cache_dir"><b>{{ param.cache_dir.label }}</b></label>
        <fast-text-field id="cache_dir" placeholder="{{ param.cache_dir.doc }}" value="${cache_dir}" style="min-width: 300px;">
        </fast-text-field>
      </div>
      <div style="display: grid;">
        <label for="shared"><b>{{ param.shared.label }}</b></label>
        <fast-checkbox id="shared" value="${shared}"></fast-checkbox>
      </div>
    </div>
    <fast-button id="preview-button" onclick="${_preview}" style="position: absolute; right: 5px; margin-top: 1.5em; z-index: 100;">
      Preview
    </fast-button>
    <fast-divider></fast-divider>
    <div id="preview-area" style="display: ${_preview_display}; margin-top: 4.5em;">
      <form id="form" role="form" style="flex: 30%; max-width: 300px; line-height: 2em;">${form}</form>
      <div id="preview" style="flex: auto; margin-left: 2em; overflow-y: auto;">${preview}</div>
    </div>
    """

    _dom_events = {'cache_dir': ['keyup']}

    def __init__(self, **params):
        super().__init__(**params)
        for name, table in self.spec.get('tables', {}).items():
            self._add_table(name=name, uri=table)

    @property
    def thumbnail(self):
        if self._thumbnail:
            return self._thumbnail
        assets = pathlib.Path(__file__).parent / 'assets'
        exts = {table.uri.split('.')[-1] for table in self.table_editors}
        if len(exts) == 1:
            filename = assets/ f'{list(exts)[0]}.png'
            if os.path.isfile(filename):
                return filename

    def _add_table(self, event=None, **kwargs):
        table = FileSourceTable(**kwargs)
        table.param.watch(self._remove_table, 'remove')
        self.table_editors += [table]

    def _remove_table(self, event):
        self.tables.remove(event.obj)
        self.param.trigger('table_editors')


class SourcesEditor(WizardItem):
    """
    Declare the data sources for your dashboard by giving your source
    a name, selecting the source type and adding the source.
    """

    disabled = param.Boolean(default=True)

    sources = param.Dict(default={}, doc="The list of sources added to the dashboard.")

    source_name = param.String(doc="Enter a name for the source")

    source_type = param.Selector(doc="Select the type of source")

    _template = """
    <span style="font-size: 2em">Source Editor</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <div style="display: flex;">
      <form role="form" style="flex: 20%; max-width: 250px; line-height: 2em;">
        <div style="display: grid;">
          <label for="source-name-${id}"><b>{{ param.source_name.label }}</b></label>
          <fast-text-field id="source-name" placeholder="{{ param.source_name.doc }}" value="${source_name}">
          </fast-text-field>
        </div>
        <div style="display: flex;">
          <div style="display: grid; flex: auto;">
            <label for="type-${id}">
              <b>{{ param.source_type.label }}</b>
            </label>
            <fast-select id="source-select" style="min-width: 150px;" value="${source_type}">
              {% for stype in param.source_type.objects %}
              <fast-option id="source-option-{{ loop.index0 }}" value="{{ stype }}">{{ stype.title() }}</fast-option>
              {% endfor %}
            </fast-select>
            <fast-tooltip anchor="type-${id}">{{ param.source_type.doc }}</fast-tooltip>
          </div>
          <fast-button id="submit" appearance="accent" style="margin-top: auto; margin-left: 1em; width: 20px;" onclick="${_add_source}" disabled="${disabled}">
            <b style="font-size: 2em;">+</b>
          </fast-button>
        </div>
      </form>
      <div id="sources" style="flex: 75%; margin-left: 1em;">
        {% for source in sources.values() %}
        <div id="source-container">${source}</div>
        <fast-divider></faster-divider>
        {% endfor %}
      </div>
    </div>
    """

    _dom_events = {'source-name': ['keyup']}

    def __init__(self, **params):
        super().__init__(**params)
        sources = param.concrete_descendents(Source)
        self.param.source_type.objects = types = [
            source.source_type for source in sources.values()
            if source.source_type is not None
        ]+['intake', 'intake_dremio']
        if self.source_type is None and types:
            self.source_type = types[0]

    @param.depends('source_name', watch=True)
    def _enable_add(self):
        self.disabled = not bool(self.source_name)

    def _add_source(self, event):
        self.spec[self.source_name] = spec = {'type': self.source_type}
        editor = SourceEditor(
            type=self.source_type, name=self.source_name, spec=spec,
            sizing_mode='stretch_width'
        )
        self.sources[self.source_name] = editor
        self.param.trigger('sources')
        self.source_name = ''
        self.ready = True



from __future__ import annotations

from typing import (
    Any, Dict, List, Optional, Type, Union,
)

import panel as pn
import param

from panel.widgets import Widget

from .base import Component
from .filters import Filter, ParamFilter, WidgetFilter
from .sources import Source
from .state import state
from .transforms import Filter as FilterTransform, SQLTransform, Transform
from .util import catch_and_notify, get_dataframe_schema
from .validation import ValidationError, match_suggestion_message


class DataFrame(param.DataFrame):
    """
    DataFrame parameter that resolves data on access.
    """

    def __get__(self, obj, objtype):
        if obj is not None and obj.__dict__.get(self._internal_name) is None:
            obj._update_data()
        return super().__get__(obj, objtype)


class Pipeline(Component):
    """
    A Pipeline represents a data Source along with any number of
    Filters and general purpose Transforms. A pipeline can be used
    to drive one or more visual outputs or leveraged as a standalone
    component to encapsulate multiple data processing steps.
    """

    data = DataFrame(doc="The current data on this source.")

    schema = param.Dict(doc="The schema of the input data.")

    auto_update = param.Boolean(default=True, constant=True, doc="""
        Whether changes in filters, transforms and references automatically
        trigger updates in the data or whether an update has to be triggered
        manually using the update event or the update button in the UI."""
    )

    source = param.ClassSelector(class_=Source, constant=True, doc="""
        The Source this pipeline is fed by."""
    )

    pipeline = param.ClassSelector(class_=None, doc="""
        Optionally a pipeline may be chained to another pipeline."""
    )

    filters = param.List(item_type=Filter, doc="""
        A list of Filters to apply to the source data."""
    )

    sql_transforms = param.List(item_type=SQLTransform, doc="""
        A list of SQLTransforms to apply to the source data."""
    )

    transforms = param.List(item_type=Transform, doc="""
        A list of Transforms to apply to the source data."""
    )

    table = param.String(doc="""
        The name of the table driving this pipeline."""
    )

    update = param.Event(label='Apply update', doc="""
        Update event trigger (if manual update is set)."""
    )

    _internal_params = ['data', 'name', 'schema']
    _required_fields = ['source']

    def __init__(self, *, source, table, **params):
        if 'schema' not in params:
            params['schema'] = source.get_schema(table)
        super().__init__(source=source, table=table, **params)
        self._init_callbacks()

    def _init_callbacks(self):
        self.param.watch(self._update_data, ['filters', 'sql_transforms', 'transforms', 'table'])
        self.source.param.watch(self._update_data, self.source._reload_params)
        for filt in self.filters:
            filt.param.watch(self._update_data, ['value'])
        for transform in self.transforms+self.sql_transforms:
            transform.param.watch(self._update_data, list(transform.param))
            for fp in transform._field_params:
                if isinstance(transform.param[fp], param.Selector):
                    transform.param[fp].objects = list(self.schema)
        if self.pipeline is not None:
            self.pipeline.param.watch(self._update_data, 'data')

    def _update_refs(self, *events):
        if self.auto_update:
            self._update_data()

    @property
    def refs(self):
        refs = self.source.refs.copy()
        for filt in self.filters:
            for ref in filt.refs:
                if ref not in refs:
                    refs.append(ref)
        for transform in self.sql_transforms+self.transforms:
            for ref in transform.refs:
                if ref not in refs:
                    refs.append(ref)
        return refs

    @param.depends('update', watch=True)
    @catch_and_notify
    def _update_data(self, *events: param.Event):
        if not self.auto_update and events and not any(
                e.name == 'update' or (e.name == 'data' and isinstance(e.obj, Pipeline)) for e in events):
            return
        query = {}

        # Compute Filter query
        for filt in self.filters:
            filt_query = filt.query
            if (filt_query is not None and not getattr(filt, 'disabled', None) and
                (filt.table is None or filt.table == self.table)):
                query[filt.field] = filt_query

        if self.pipeline is None:
            # Compute SQL transform expression
            if self.sql_transforms:
                if not self.source._supports_sql:
                    raise ValidationError(
                        'Can only use sql transforms source that support them. '
                        f'Found source typed {self.source.source_type!r} instead.'
                    )
                query['sql_transforms'] = self.sql_transforms
            data = self.source.get(self.table, **query)
        else:
            if self.pipeline.data is None:
                self.pipeline._update_data()
            data = FilterTransform.apply_to(
                self.pipeline.data, conditions=list(query.items())
            )

        # Apply ParamFilter
        for filt in self.filters:
            if not isinstance(filt, ParamFilter):
                continue
            from holoviews import Dataset
            if filt.value is not None:
                ds = Dataset(data)
                data = ds.select(filt.value).data

        # Apply transforms
        for transform in self.transforms:
            data = transform.apply(data)

        self.data = data

    @classmethod
    def _validate_source(cls, source_spec, spec, context):
        if isinstance(source_spec, str):
            if source_spec not in context['sources']:
                msg = f'Pipeline specified non-existent source {source_spec!r}.'
                msg = match_suggestion_message(source_spec, list(context['sources']), msg)
                raise ValidationError(msg, spec, source_spec)
            return source_spec
        return Source.validate(source_spec, context)

    @classmethod
    def _validate_filters(cls, filter_specs, spec, context):
        for filter_spec in (filter_specs if isinstance(filter_specs, list) else filter_specs.values()):
            if not isinstance(filter_spec, str):
                continue
            elif not isinstance(spec['source'], str):
                raise ValidationError(
                    'Pipeline may only reference filters by name if the Source has been '
                    'defined by reference. Please simply move the filter definition from '
                    'the Source to the Pipeline.', spec, filter_spec
                )
            source = context['sources'][spec['source']]
            if 'filters' not in source:
                raise ValidationError(
                    f'Pipeline could not resolve {filter_spec!r} filter on {spec["source"]} source, '
                    'the source does not define any filters. ', spec, filter_spec
                )
            elif filter_spec not in source['filters']:
                msg = f'Pipeline could not resolve {filter_spec!r} filter on {spec["source"]} source.'
                msg = match_suggestion_message(filter_spec, list(source['filters']), msg)
                raise ValidationError(msg, spec, filter_spec)
        return cls._validate_dict_or_list_subtypes('filters', Filter, filter_specs, spec, context)

    @classmethod
    def validate(cls, spec, context=None):
        if isinstance(spec, str):
            if spec not in context['pipelines']:
                msg = f'Referenced non-existent pipeline {spec!r}.'
                msg = match_suggestion_message(spec, list(context['pipelines']), msg)
                raise ValidationError(msg, spec, spec)
            return spec
        return super().validate(spec, context)

    @classmethod
    def from_spec(
        cls, spec: Dict[str, Any] | str, source: Optional[Source] = None,
        source_filters: Optional[List[Filter]] = None
    ):
        if isinstance(spec, str):
            return state.pipelines[spec]

        spec = spec.copy()
        if source is not None:
            spec['source'] = source
        params = dict(spec)

        # Resolve source
        source = spec['source']
        if isinstance(source, dict):
            source = Source.from_spec(source)
        elif isinstance(source, str):
            if source in state.sources:
                source = state.sources[source]
            else:
                source = state.load_source(source, state.spec['sources'][source])
        params['source'] = source

        # Validate table
        table = params.get('table')
        tables = source.get_tables()
        if table is None:
            if len(tables) > 1:
                raise ValidationError(
                    "The Pipeline specification does not contain a table and the "
                    "supplied Source has multiple tables. Please specify one of the "
                    f"following tables: {tables} in the pipeline specification.",
                    spec
                )
            params['table'] = table = tables[0]
        elif table not in tables:
            raise ValidationError(
                "The Pipeline specification references a table that is not "
                "available on the specified source. ", spec
            )

        # Resolve filters
        params['filters'] = filters = []
        filter_specs = spec.pop('filters', {})
        if filter_specs:
            params['schema'] = schema = source.get_schema(table)
        for filt_spec in (filter_specs.items() if isinstance(filter_specs, dict) else filter_specs):
            if isinstance(filt_spec, tuple):
                filt_spec = dict(filt_spec[1], table=table, name=filt_spec[0])
            else:
                filt_spec = dict(filt_spec, table=table)
            filt = Filter.from_spec(filt_spec, {table: schema}, source_filters)
            filters.append(filt)

        # Resolve transforms
        transform_specs = spec.pop('transforms', [])
        params['transforms'] = [Transform.from_spec(tspec) for tspec in transform_specs]
        sql_transform_specs = spec.pop('sql_transforms', [])
        params['sql_transforms'] = [Transform.from_spec(tspec) for tspec in sql_transform_specs]
        return cls(**params)

    def add_filter(self, filt: Union[Filter, Type[Filter], Widget], field: Optional[str] = None, **kwargs):
        """
        Add a filter to the pipeline.

        Arguments
        ---------
        filt: Filter | Type[Filter]
           The filter instance or filter type to add.
        field: str | None
           The field to filter on (required to instantiate Filter type).
        """
        if isinstance(filt, str):
            filt = Filter._get_type(filt)
        elif isinstance(filt, Widget):
            filt = WidgetFilter(
                widget=filt, field=field, table=self.table,
                schema={self.table: self.schema}
            )
        if not isinstance(filt, Filter):
            tspec = f'{filt.__module__}.{filt.__name__}'
            filt = Filter.from_spec(
                dict({'type': tspec, 'field': field, 'table': self.table}, **kwargs),
                {self.table: self.schema}
            )
        self.filters.append(filt)
        filt.param.watch(self._update_data, ['value'])
        self._update_data()

    def add_transform(self, transform: Transform, **kwargs):
        """
        Add a (SQL)Transform to the pipeline.

        Arguments
        ---------
        filt: Transform
           The Transform instance to add.
        """
        if isinstance(transform, str):
            transform = Transform._get_type(transform)(**kwargs)
        if isinstance(transform, SQLTransform):
            self.sql_transforms.append(transform)
        else:
            self.transforms.append(transform)
        fields = list(self.schema)
        for fparam in transform._field_params:
            transform.param[fparam].objects = fields
            transform.param.update(**{fparam: kwargs.get(fparam, fields)})
        transform.param.watch(self._update_data, list(transform.param))
        self._update_data()

    def chain(
        self,
        filters: Optional[List[Filter]]=None,
        transforms: Optional[List[Transform]] = None,
        sql_transforms: Optional[List[Transform]] = None
    ):
        """
        Chains additional filtering, transform and sql_transform operations
        on an existing pipeline. Note that if one or more sql_transforms
        are provided the pipeline is cloned rather than applying the
        operations on top of the existing pipeline.

        Arguments
        ---------
        filters: List[Filter] | None
          Additional filters to apply on top of existing pipeline.
        transforms: List[Transform] | None
          Additional transforms to apply on top of existing pipeline.
        sql_transforms: List[SQLTransform] | None
          Additional filters to apply on top of existing pipeline.

        Returns
        -------
        Pipeline
        """
        if sql_transforms:
            params = {
                'filters': self.filters + (filters or []),
                'transforms': self.transforms + (transforms or []),
                'sql_transforms': self.sql_transforms + (sql_transforms or []),
                'data': None
            }
        else:
            params = {
                'filters': filters or [],
                'transforms': transforms or [],
                'sql_transforms': [],
                'pipeline': self,
                'schema': get_dataframe_schema(self.data)['items']['properties'],
                'data': None
            }
        return self.clone(**params)

    def clone(self, **params) -> Pipeline:
        """
        Create a new instance of the pipeline with optionally overridden parameter values.
        """
        return type(self)(**dict({p: v for p, v in self.param.values().items()
                                  if p != 'name'}, **params))

    def traverse(self, type) -> List[Transform] | List[Filter]:
        """
        Returns all Filter or Transform objects in a potentially chained
        pipeline.
        """
        if type not in ('filters', 'transforms', 'sql_transforms'):
            raise TypeError(f'May only traverse Pipeline filters, transforms or sql_transforms, not {type}')
        objects = []
        pipeline = self
        while pipeline is not None:
            objects.extend(getattr(self, type))
            pipeline = pipeline.pipeline
        return objects

    @property
    def control_panel(self) -> pn.Column:
        col = pn.Column()
        filters = [filt.panel for filt in self.filters]
        if any(filters):
            col.append('<div style="font-size: 1.5em; font-weight: bold;">Filters</div>')
        col.extend([filt for filt in filters if filt is not None])
        transforms = [t.control_panel for t in self.transforms+self.sql_transforms if t.controls]
        if transforms:
            col.append('<div style="font-size: 1.5em; font-weight: bold;">Transforms</div>')
        col.extend(transforms)
        if not self.auto_update:
            col.append(self.param['update'])
        return col


Pipeline.param.pipeline.class_ = Pipeline



"""
The Transform components allow transforming tables in arbitrary ways.
"""

import datetime as dt

import numpy as np
import pandas as pd
import panel as pn
import param

from ..base import MultiTypeComponent
from ..state import state
from ..util import is_ref


class Transform(MultiTypeComponent):
    """
    A Transform provides the ability to transform a table supplied by
    a Source.
    """

    controls = param.List(default=[], doc="""
        Parameters that should be exposed as widgets in the UI.""")

    transform_type = None

    _field_params = []

    __abstract = True

    @classmethod
    def from_spec(cls, spec):
        """
        Resolves a Transform specification.

        Parameters
        ----------
        spec: dict
            Specification declared as a dictionary of parameter values.

        Returns
        -------
        The resolved Transform object.
        """
        spec = dict(spec)
        transform_type = Transform._get_type(spec.pop('type', None))
        new_spec, refs = {}, {}
        for k, v in spec.items():
            if (k in transform_type.param and
                isinstance(transform_type.param[k], param.ListSelector) and
                not isinstance(v, list)):
                v = [v]
            if is_ref(v):
                refs[k] = v
                v = state.resolve_reference(v)
            new_spec[k] = v

        # Resolve any specs for the controls
        controls, control_kwargs = [], {}
        for control in new_spec.get('controls', []):
            if isinstance(control, dict):
                ckws = {}
                if 'options' in control:
                    options = control['options']
                    if isinstance(options, str):
                        options = state.resolve_reference(options)
                    ckws['objects'] = options
                if 'start' in control or 'end' in control:
                    ckws['bounds'] = (control.get('start'), control.get('end'))
                control = control['name']
                control_kwargs[control] = ckws
            controls.append(control)
        new_spec['controls'] = controls

        # Instantiate the transform
        transform = transform_type(refs=refs, **new_spec)

        # Modify the parameters for the controls
        for p, vs in control_kwargs.items():
            p = transform.param[p]
            for attr, val in vs.items():
                if hasattr(p, attr):
                    setattr(p, attr, val)
                else:
                    attr = 'options' if attr == 'objects' else attr
                    cls.param.warning(
                        f"{transform_type.__name__} is of type {type(p).__name__} "
                        f"and has not attribute {attr!r}. Ensure the controls "
                        "parameter supports the provided options, e.g. if "
                        "you are declaring 'options' ensure that the parameter "
                        "is a param.Selector type."
                    )
        return transform

    @classmethod
    def apply_to(cls, table, **kwargs):
        """
        Calls the apply method based on keyword arguments passed to define transform.

        Parameters
        ----------
        table: `pandas.DataFrame`

        Returns
        -------
        A DataFrame with the results of the transformation.
        """
        return cls(**kwargs).apply(table)

    def apply(self, table):
        """
        Given a table transform it in some way and return it.

        Parameters
        ----------
        table : DataFrame
            The queried table as a DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame containing the transformed data.
        """
        return table

    @property
    def control_panel(self):
        return pn.Param(
            self.param, parameters=self.controls, sizing_mode='stretch_width',
            margin=(-10, 0, 5, 0)
        )

    @property
    def refs(self):
        refs = super().refs
        for c in self.controls:
            if c not in refs:
                refs.append(c)
        return refs


class Filter(Transform):
    """
    Transform that applies the query values from Lumen Filter
    components to a dataframe. The filter values can be one of the
    following:

      * scalar: A scalar value will be matched using equality operators
      * tuple:  A tuple value specifies a numeric or date range.
      * list:   A list value specifies a set of categories to match against.
      * list(tuple): A list of tuples specifies a list of ranges.
    """

    conditions = param.List(doc="""
      List of filter conditions expressed as tuples of the column
      name and the filter value.""")

    @classmethod
    def _range_filter(cls, column, start, end):
        if column.dtype.kind == 'M':
            if isinstance(start, dt.date) and not isinstance(start, dt.datetime):
                start = dt.datetime(*start.timetuple()[:3], 0, 0, 0)
            if isinstance(end, dt.date) and not isinstance(end, dt.datetime):
                end = dt.datetime(*end.timetuple()[:3], 23, 59, 59)
        if start is None and end is None:
            return None
        elif start is None:
            mask = column<=end
        elif end is None:
            mask = column>=start
        else:
            mask = (column>=start) & (column<=end)
        return mask

    def apply(self, df):
        filters = []
        for k, val in self.conditions:
            if k not in df.columns:
                continue
            column = df[k]
            if np.isscalar(val) or isinstance(val, dt.date):
                if (column.dtype.kind == 'M' and isinstance(val, dt.date)
                    and not isinstance(val, dt.datetime)):
                    val = dt.datetime(*val.timetuple()[:3], 0, 0, 0)
                mask = column == val
            elif isinstance(val, list) and all(isinstance(v, tuple) and len(v) == 2 for v in val):
                val = [v for v in val if v is not None]
                if not val:
                    continue
                mask = self._range_filter(column, *val[0])
                for v in val[1:]:
                    mask |= self._range_filter(column, *v)
                if mask is not None:
                    filters.append(mask)
                continue
            elif isinstance(val, list):
                if not val:
                    continue
                mask = column.isin(val)
            elif isinstance(val, tuple):
                mask = self._range_filter(column, *val)
            else:
                self.param.warning(
                    'Condition {val!r} on {col!r} column not understood. '
                    'Filter query will not be applied.'
                )
                continue
            if mask is not None:
                filters.append(mask)
        if filters:
            mask = filters[0]
            for f in filters[1:]:
                mask &= f
            df = df[mask]
        return df


class HistoryTransform(Transform):
    """
    The HistoryTransform accumulates a history of the queried data in
    a buffer up to the supplied length and (optionally) adds a
    date_column to the data.
    """

    date_column = param.Selector(doc="""
        If defined adds a date column with the supplied name.""")

    length = param.Integer(default=10, bounds=(1, None), doc="""
        Accumulates a history of data.""")

    transform_type = 'history'

    _field_params = ['date_column']

    def __init__(self, **params):
        super().__init__(**params)
        self._buffer = []

    def apply(self, table):
        """
        Accumulates a history of the data in a buffer up to the
        declared `length` and optionally adds the current datetime to
        the declared `date_column`.

        Parameters
        ----------
        data : DataFrame
            The queried table as a DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame containing the buffered history of the data.
        """
        if self.date_column:
            table = table.copy()
            table[self.date_column] = dt.datetime.now()
        self._buffer.append(table)
        self._buffer[:] = self._buffer[-self.length:]
        return pd.concat(self._buffer)


class Aggregate(Transform):
    """
    Aggregate one or more columns or indexes, see `pandas.DataFrame.groupby`.

    df.groupby(<by>)[<columns>].<method>()[.reset_index()]
    """

    by = param.ListSelector(doc="""
        Columns or indexes to group by.""")

    columns = param.ListSelector(doc="""
        Columns to aggregate.""")

    with_index = param.Boolean(default=True, doc="""
        Whether to make the groupby columns indexes.""")

    method = param.String(default=None, doc="""
        Name of aggregation method.""")

    kwargs = param.Dict(default={}, doc="""
        Keyword arguments to the aggregation method.""")

    transform_type = 'aggregate'

    _field_params = ['by', 'columns']

    def apply(self, table):
        grouped = table.groupby(self.by)
        if self.columns:
            grouped = grouped[self.columns]
        agg = getattr(grouped, self.method)(**self.kwargs)
        return agg if self.with_index else agg.reset_index()


class Sort(Transform):
    """
    Sort on one or more columns, see `pandas.DataFrame.sort_values`.

    df.sort_values(<by>, ascending=<ascending>)
    """

    by = param.ListSelector(default=[], doc="""
       Columns or indexes to sort by.""")

    ascending = param.ClassSelector(default=True, class_=(bool, list), doc="""
       Sort ascending vs. descending. Specify list for multiple sort
       orders. If this is a list of bools, must match the length of
       the by.""")

    transform_type = 'sort'

    _field_params = ['by']

    def apply(self, table):
        return table.sort_values(self.by, ascending=self.ascending)


class Query(Transform):
    """
    Applies the `pandas.DataFrame.query` method.

    df.query(<query>)
    """

    query = param.String(doc="""
        The query to apply to the table.""")

    transform_type = 'query'

    def apply(self, table):
        return table.query(self.query)


class Columns(Transform):
    """
    Selects a subset of columns.

    df[<columns>]
    """

    columns = param.ListSelector(doc="""
        The subset of columns to select.""")

    transform_type = 'columns'

    _field_params = ['columns']

    def apply(self, table):
        return table[self.columns]


class Astype(Transform):
    """
    Transforms the type of one or more columns.
    """

    dtypes = param.Dict(doc="Mapping from column name to new type.")

    transform_type = 'as_type'

    def apply(self, table):
        table = table.copy()
        for col, dtype in self.dtypes.items():
            table[col] = table[col].astype(dtype)
        return table


class Stack(Transform):
    """
    Stacks the declared level, see `pandas.DataFrame.stack`.

    df.stack(<level>)
    """

    dropna = param.Boolean(default=True, doc="""
        Whether to drop rows in the resulting Frame/Series with missing values.
        Stacking a column level onto the index axis can create combinations of
        index and column values that are missing from the original
        dataframe.""")

    level = param.ClassSelector(default=-1, class_=(int, list, str), doc="""
        The indexes to stack.""")

    transform_type = 'stack'

    def apply(self, table):
        return table.stack(level=self.level, dropna=self.dropna)


class Unstack(Transform):
    """
    Unstacks the declared level(s), see `pandas.DataFrame.stack`.

    df.unstack(<level>)
    """

    fill_value = param.ClassSelector(default=None, class_=(int, str, dict), doc="""
        Replace NaN with this value if the unstack produces missing values.""")

    level = param.ClassSelector(default=-1, class_=(int, list, str), doc="""
        The indexes to unstack.""")

    transform_type = 'unstack'

    def apply(self, table):
        return table.unstack(level=self.level, fill_value=self.fill_value)


class Iloc(Transform):
    """
    Applies integer slicing to the data, see `pandas.DataFrame.iloc`.

    df.iloc[<start>:<end>]
    """

    start = param.Integer(default=None)

    end = param.Integer(default=None)

    transform_type = 'iloc'

    def apply(self, table):
        return table.iloc[self.start:self.end]


class Sample(Transform):
    """
    Random sample of items.

    df.sample(n=<n>, frac=<frac>, replace=<replace>)
    """

    n = param.Integer(default=None, doc="""
        Number of items to return.""")

    frac = param.Number(default=None, bounds=(0, 1), doc="""
        Fraction of axis items to return.""")

    replace = param.Boolean(default=False, doc="""
        Sample with or without replacement.""")

    transform_type = 'sample'

    def apply(self, table):
        return table.sample(n=self.n, frac=self.frac, replace=self.replace)


class Compute(Transform):
    """
    Turns a Dask DataFrame into a pandas DataFrame.
    """

    transform_type = 'compute'

    def apply(self, table):
        return table.compute()

class Pivot(Transform):
    """
    Pivots a DataFrame given an index, columns, and values.
    """

    index = param.String(default=None, doc="""
        Column to use to make new frame's index.
        If None, uses existing index.""")

    columns = param.String(default=None, doc="""
        Column to use to make new frame’s columns.""")

    values = param.ListSelector(default=None, doc="""
        Column(s) to use for populating new frame’s values.
        If not specified, all remaining columns will be used
        and the result will have hierarchically indexed columns.""")

    transform_type = 'pivot'

    def apply(self, table):
        return table.pivot(index=self.index, columns=self.columns, values=self.values)


class Melt(Transform):
    """
    Melts a DataFrame given the id_vars and value_vars.
    """

    id_vars = param.ListSelector(default=[], doc="""
        Column(s) to use as identifier variables.""")

    ignore_index = param.Boolean(default=True, doc="""
        If True, original index is ignored. If False, the original
        index is retained. Index labels will be repeated as
        necessary.""")

    value_vars = param.ListSelector(default=None, doc="""
        Column(s) to unpivot. If not specified, uses all columns that
        are not set as `id_vars`.""")

    var_name = param.String(default=None, doc="""
         Name to use for the 'variable' column. If None it uses
         ``frame.columns.name`` or 'variable'.""")

    value_name = param.String(default='value', doc="""
         Name to use for the 'value' column.""")

    transform_type = 'melt'

    _field_params = ['id_vars', 'value_vars']

    def apply(self, table):
        return pd.melt(table, id_vars=self.id_vars, value_vars=self.value_vars,
                       var_name=self.var_name, value_name=self.value_name,
                       ignore_index=self.ignore_index)


class SetIndex(Transform):
    """
    Set the DataFrame index using existing columns, see `pandas.DataFrame.set_index`.

    df.set_index(<keys>, drop=<drop>, append=<append>, verify_integrity=<verify_integrity>)
    """

    append = param.Boolean(default=False, doc="""
        Whether to append columns to existing index.""")

    drop = param.Boolean(default=True, doc="""
        Delete columns to be used as the new index.""")

    keys = param.ClassSelector(default=None, class_=(str, list), doc="""
        This parameter can be either a single column key or a list
        containing column keys.""")

    verify_integrity = param.Boolean(default=False, doc="""
        Check the new index for duplicates. Otherwise defer the check
        until necessary. Setting to False will improve the performance
        of this method.""")

    transform_type = 'set_index'

    _field_params = ['keys']

    def apply(self, table):
        return table.set_index(
            self.keys, drop=self.drop, append=self.append,
            verify_integrity=self.verify_integrity
        )


class ResetIndex(Transform):
    """
    Sort on one or more columns, see `pandas.DataFrame.sort_values`.

    df.sort_values(<by>, ascending=<ascending>)
    """

    col_fill = param.String(default="", doc="""
        If the columns have multiple levels, determines how the other
        levels are named. If None then the index name is repeated.""")

    col_level = param.ClassSelector(default=0, class_=(int, str), doc="""
        If the columns have multiple levels, determines which level the
        labels are inserted into. By default it is inserted into the
        first level.""")

    drop = param.Boolean(default=False, doc="""
        Do not try to insert index into dataframe columns. This resets
        the index to the default integer index.""")

    level = param.ClassSelector(default=None, class_=(int, str, list), doc="""
        Only remove the given levels from the index. Removes all levels
        by default.""")

    transform_type = 'reset_index'

    def apply(self, table):
        return table.reset_index(
            drop=self.drop, col_fill=self.col_fill, col_level=self.col_level,
            level=self.level
        )


class Rename(Transform):
    """
    Alter axes labels,  see `pandas.DataFrame.rename`.

    df.rename(mapper=<mapper>, columns=<columns>, index=<index>,
              level=<level>, axis=<axis>, copy=<copy>)
    """

    axis = param.ClassSelector(default=0, class_=(int, str), doc="""
        The axis to rename. 0 or 'index', 1 or 'columns'""")

    columns = param.Dict(default=None, doc="""
        Alternative to specifying axis (`mapper, axis=1` is equivalent to
        `columns=mapper`).""")

    copy = param.Boolean(default=True, doc="""
        Also copy underlying data.""")

    index = param.Dict(default=None, doc="""
        Alternative to specifying axis (`mapper, axis=0` is equivalent to
        `index=mapper`).""")

    mapper = param.Dict(default=None, doc="""
        Dict to apply to that axis' values. Use either `mapper` and `axis` to
        specify the axis to target with `mapper`, or `index` and `columns`.""")

    level = param.ClassSelector(default=None, class_=(int, str), doc="""
        In case of a MultiIndex, only rename labels in the specified level.""")

    transform_type = 'rename'

    def apply(self, table):
        return table.rename(
            axis=self.axis, columns=self.columns, copy=self.copy,
            index=self.index, mapper=self.mapper, level=self.level,
        )


class RenameAxis(Transform):
    """
    Set the name of the axis for the index or columns,
    see `pandas.DataFrame.rename_axis`.

    df.rename_axis(mapper=<mapper>, columns=<columns>, index=<index>,
                  axis=<axis>, copy=<copy>)
    """

    axis = param.ClassSelector(default=0, class_=(int, str), doc="""
        The axis to rename. 0 or 'index', 1 or 'columns'""")

    columns = param.ClassSelector(default=None, class_=(str, list, dict), doc="""
        A scalar, list-like, dict-like to apply to that axis' values.
        Note that the columns parameter is not allowed if the object
        is a Series. This parameter only apply for DataFrame type objects.
        Use either mapper and axis to specify the axis to target with
        mapper, or index and/or columns.""")

    copy = param.Boolean(default=True, doc="""
        Also copy underlying data.""")

    index = param.ClassSelector(default=None, class_=(str, list, dict), doc="""
        A scalar, list-like, dict-like to apply to that axis' values.
        Note that the columns parameter is not allowed if the object
        is a Series. This parameter only apply for DataFrame type objects.
        Use either mapper and axis to specify the axis to target with
        mapper, or index and/or columns.""")

    mapper = param.ClassSelector(default=None, class_=(str, list), doc="""
        Value to set the axis name attribute.""")

    transform_type = 'rename_axis'

    def apply(self, table):
        return table.rename_axis(
            axis=self.axis, columns=self.columns, copy=self.copy,
            index=self.index, mapper=self.mapper,
        )


class project_lnglat(Transform):
    """
    Projects the given (longitude, latitude) values into Web Mercator
    coordinates (meters East of Greenwich and meters North of the Equator).
    """

    longitude = param.String(default='longitude', doc="Longitude column")
    latitude = param.String(default='longitude', doc="Latitude column")

    transform_type = 'project_lnglat'

    def apply(self, table):
        table = table.copy()
        longitude = table[self.longitude]
        latitude = table[self.latitude]

        origin_shift = np.pi * 6378137
        table[self.longitude] = longitude * origin_shift / 180.0
        table[self.latitude] = np.log(np.tan((90 + latitude) * np.pi / 360.0)) * origin_shift / np.pi
        return table



import panel as pn
import param

from panel.reactive import ReactiveHTML

from lumen.state import state as lm_state
from lumen.util import catch_and_notify
from lumen.views import View

from .base import WizardItem
from .gallery import Gallery, GalleryItem
from .sources import ASSETS_DIR
from .state import state


class ViewsEditor(WizardItem):
    """
    Declare the views for your dashboard.
    """

    spec = param.List(default=[], precedence=-1)

    pipeline  = param.String()

    view_type = param.Selector()

    pipelines = param.List(label='Select a pipeline')

    views = param.List()

    _template = """
    <span style="font-size: 2em">View Editor</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <div style="display: flex;">
      <form role="form" style="flex: 30%; max-width: 300px; line-height: 2em;">
        <div style="display: grid; flex: auto;">
          <label for="view-select-${id}"><b>{{ param.view_type.label }}</b></label>
          <fast-select id="view-select" style="min-width: 150px;" value="${view_type}">
          {% for vtype in param.view_type.objects %}
            <fast-option value="{{ vtype }}">{{ vtype.title() }}</fast-option>
          {% endfor %}
          </fast-select>
          <fast-tooltip anchor="view-select-${id}">{{ param.view_type.doc }}</fast-tooltip>
        </div>
        <div style="display: flex;">
          <div style="display: grid;">
            <label for="pipelines-${id}"><b>{{ param.pipelines.label }}</b></label>
            <fast-select id="pipelines" style="max-width: 300px; min-width: 240px;" value="${pipeline}">
            {% for ppl in pipelines %}
              <fast-option value="{{ ppl|tojson }}">{{ ppl }}</fast-option>
            {% endfor %}
            </fast-select>
            <fast-tooltip anchor="pipelines-${id}">{{ param.pipelines.doc }}</fast-tooltip>
          </div>
          <fast-button id="submit" appearance="accent" style="margin-top: auto; margin-left: 1em; width: 20px;" onclick="${_add_view}">
            <b style="font-size: 2em;">+</b>
          </fast-button>
        </div>
      </form>
      <div id="view-list" style="flex: auto; margin-left: 2em;">
        {% for view in views %}
        <div id="view-container">${view}</div>
        <fast-divider></faster-divider>
        {% endfor %}
      </div>
    </div>
    """

    def __init__(self, **params):
        super().__init__(**params)
        views = param.concrete_descendents(View)
        self.param.view_type.objects = [
            view.view_type for view in views.values() if view.view_type
        ]
        state.pipelines.param.watch(self._update_pipelines, 'pipelines')

    def _add_view(self, event):
        editor = ViewEditor(
            view_type=self.view_type, pipeline=self.pipeline,
            spec={'pipeline': self.pipeline, 'type': self.view_type}
        )
        editor.param.watch(self._remove_view, 'remove')
        editor.render()
        self.views.append(editor)
        self.param.trigger('views')

    @catch_and_notify
    def _update_pipelines(self, event):
        self.pipelines = list(event.new)
        if not self.pipeline and self.pipelines:
            self.pipeline = self.pipelines[0]

    def _remove_view(self, event):
        self.views.remove(event.obj)
        self.param.trigger('views')


class ViewEditor(ReactiveHTML):

    # Specification parameters

    spec = param.Dict(default={})

    sizing_mode = param.String(default='stretch_both')

    pipeline = param.Parameter(precedence=-1)

    view_type = param.String(default='None')

    # Display parameters

    remove = param.Boolean(default=False)

    view = param.Parameter()

    _template = """
    <span style="font-size: 2em">{{ view_type.title() }} Editor</span>
    <p>Configure the view.</p>
    <div style="width: 300px; display: grid; margin-right: 1em;">
      <label for="view-name"><b>View Name</b></label>
      <fast-text-field id="view-name" placeholder="Enter a name" value="${name}"></fast-text-field>
    </div>
    <fast-divider></fast-divider>
    <div id='view'>${view}</div>
    """

    _dom_events = {'view-name': ['keyup']}

    def __new__(cls, **params):
        if cls is not ViewEditor:
            return super().__new__(cls)
        editors = param.concrete_descendents(cls)
        view_type = params.get('spec', {}).get('type')
        for editor in editors.values():
            if editor.view_type == view_type:
                return super().__new__(editor)
        return super().__new__(cls)

    def __init__(self, **params):
        spec = params.pop('spec', {})
        params.update(**{
            k: v for k, v in spec.items() if k in self.param and k not in params
        })
        super().__init__(spec=spec, **params)

    @property
    def description(self):
        return (
            f"A {self.view_type} view of the {self.spec['pipeline']!r} source."
        )

    def render(self):
        pass

    @property
    def thumbnail(self):
        return ASSETS_DIR / 'view.png'

    def _remove(self, event):
        self.remove = True

    def _update_spec(self, *events):
        for event in events:
            if not isinstance(event.new, param.Parameterized):
                self.spec[event.name] = event.new


class ViewGalleryItem(GalleryItem):

    editor = param.ClassSelector(class_=ViewEditor, precedence=-1)

    selected = param.Boolean(default=False, doc="""
        Whether the item has been selected.""")

    def __init__(self, **params):
        super().__init__(**params)
        self.thumbnail = self.editor.thumbnail
        self.description = self.editor.description
        self.view = pn.pane.PNG(self.thumbnail, height=200, align='center')
        self._modal_content = [self.editor]

    def _open_modal(self, event):
        self.editor.render()
        super()._open_modal(event)


class ViewGallery(WizardItem, Gallery):
    """
    Select the views to add to your dashboard specification.
    """

    path = param.Foldername()

    spec = param.List()

    views = param.List(default=[], precedence=-1, doc="""
        The list of views added to the dashboard.""")

    _template = """
    <span style="font-size: 1.5em">Views</span>
    <fast-divider></fast-divider>
    <span style="font-size: 1.2em; font-weight: bold;">{{ __doc__ }}</p>
    <div id="items" style="margin: 1em 0; display: flex; flex-wrap: wrap; gap: 1em;">
    {% for item in items %}
      <fast-card id="view-container" class="gallery-item" style="width: 350px; height: 400px;">
        ${item}
      </fast-card>
    {% endfor %}
      <fast-card id="view-container-new" class="gallery-item" style="height: 400px; width: 350px; padding: 1em;">
        <div style="display: grid;">
          <span style="font-size: 1.25em; font-weight: bold;">Add new view</span>
          <i id="add-button" onclick="${_create_new}" class="fa fa-plus" style="font-size: 14em; margin: 0.2em auto;" aria-hidden="true"></i>
        </div>
      </fast-card>
    </div>
    """

    _editor_type = ViewEditor

    _gallery_item = ViewGalleryItem

    def __init__(self, **params):
        super().__init__(**params)
        self._items = self.items
        self._editor = ViewsEditor(spec=self.spec, margin=10)
        self._save_button = pn.widgets.Button(name='Save view')
        self._save_button.on_click(self._save_view)
        self._watchers = {}
        state.pipelines.param.watch(self._update_items, 'items')
        self._update_items()

    def _create_new(self, event):
        if state.modal.objects == [self._editor, self._save_button]:
            state.template.open_modal()
            return
        state.modal.loading = True
        state.template.open_modal()
        state.modal[:] = [self._editor, self._save_button]
        state.modal.loading = False

    def _update_items(self, *events):
        for name, item in state.pipelines.items.items():
            if name not in self._watchers:
                self._watchers[name] = item.param.watch(self._update_items, 'selected')
        for name, item in self.items.items():
            if name not in self._items:
                self._items[name] = item
        self.items = {
            name: item for name, item in self._items.items()
            if item.spec['pipeline'] in lm_state.pipelines
        }
        self.param.trigger('items')

    def _save_view(self, event):
        for view in self._editor.views:
            self.items[view.name] = ViewGalleryItem(
                name=view.name, spec=view.spec, selected=True, editor=view,
                thumbnail=view.thumbnail
            )
            self.views.append(view)
        self.param.trigger('views')
        self.param.trigger('items')
        self._editor.views = []
        state.template.close_modal()

    @param.depends('spec', watch=True)
    def _update_params(self):
        for name, spec in self.spec.items():
            self.views = editor = ViewEditor(name=name, spec=spec)
            self.items[name] = ViewGalleryItem(
                name=name, spec=spec, selected=True, editor=editor,
                thumbnail=editor.thumbnail
            )
        self.param.trigger('views')
        self.param.trigger('items')


class TableViewEditor(ViewEditor):

    page_size = param.Integer(default=15, doc="Declare the page size")

    view_type = param.String(default='table')

    _template = """
    <span style="font-size: 2em">{{ view_type.title() }} Editor</span>
    <p>Configure the view.</p>
    <div style="width: 300px; display: grid; margin-right: 1em;">
      <label for="view-name"><b>View Name</b></label>
      <fast-text-field id="view-name" placeholder="Enter a name" value="${name}"></fast-text-field>
    </div>
    <fast-divider></fast-divider>
    <div id="view">${view}</div>
    """

    _dom_events = {'view-name': ['keyup']}

    def render(self):
        kwargs = dict(self.spec)
        pipeline = lm_state.pipelines[kwargs.pop('pipeline')]
        if 'sizing_mode' not in kwargs:
            kwargs['sizing_mode'] = 'stretch_width'
        self.tabulator = pn.widgets.Tabulator(
            pipeline.data, pagination='remote', page_size=12, height=400, **kwargs
        )
        controls = ['theme', 'layout', 'page_size', 'pagination']
        control_widgets = self.tabulator.controls(
            controls, margin=(0, 20, 0, 0), jslink=False
        )
        control_widgets.width = 300
        for w in control_widgets:
            w.width = 250
        self.view = pn.Row(
            control_widgets,
            self.tabulator,
            sizing_mode='stretch_width'
        )
        self.tabulator.param.watch(self._update_spec, controls)
        self.tabulator.param.trigger(*controls)

    @property
    def thumbnail(self):
        return ASSETS_DIR / 'tabulator.png'


class PerspectiveViewEditor(ViewEditor):

    view_type = param.String(default='perspective')

    _defaults = dict(sizing_mode='stretch_both', min_height=500, theme='material-dark')

    def __init__(self, **params):
        super().__init__(**{k: v for k, v in params.items() if k in self.param})

    def render(self):
        kwargs = dict(self.spec)
        pipeline = lm_state.pipelines[kwargs.pop('pipeline')]
        self.view = pn.pane.Perspective(pipeline, **dict(self._defaults, **kwargs))
        self.view.param.watch(self._update_spec, list(self.view.param))

    @property
    def thumbnail(self):
        return ASSETS_DIR / 'perspective.png'


class hvPlotViewEditor(ViewEditor):

    selection_group = param.String(default=None, allow_None=True, precedence=-1)

    view_type = param.String(default='hvplot')

    def __init__(self, **params):
        import hvplot.pandas  # noqa
        super().__init__(**params)

    def render(self):
        from hvplot.ui import hvDataFrameExplorer
        kwargs = dict(self.spec)
        del kwargs['type']
        pipeline = lm_state.pipelines[kwargs.pop('pipeline')]
        self.view = hvDataFrameExplorer(pipeline.data, **kwargs)
        self.view.param.watch(self._update_spec, list(self.view.param))
        self.view.axes.param.watch(self._update_spec, list(self.view.axes.param))
        self.view.operations.param.watch(self._update_spec, list(self.view.operations.param))
        self.view.style.param.watch(self._update_spec, list(self.view.style.param))
        self.spec['x'] = self.view.x
        self.spec['y'] = self.view.y
        self.spec['kind'] = self.view.kind

    @property
    def thumbnail(self):
        return ASSETS_DIR / 'hvplot.png'

    def _update_spec(self, *events):
        pipeline = self.spec['pipeline']
        self.spec.clear()
        self.spec['type'] = self.view_type
        self.spec['pipeline'] = pipeline
        self.spec.update(self.view.settings())



import panel as pn
import param

from panel.reactive import ReactiveHTML

from .fast import FastDivider


class Wizard(ReactiveHTML):

    items = param.List(precedence=-1)

    current = param.Parameter()

    previous_disable = param.Boolean(True)

    next_disable = param.Boolean(True)

    spec = param.Dict(precedence=-1)

    _template = """
    <div id="wizard-content" style="display: flex; flex-direction: column; justify-content: space-between; height: 100%;">
    <div id="wizard" style="overflow: clip auto; padding-right: 1em;">${current}</div>
    <div id="wizard-footer" style="margin-top: auto;">
      <fast-divider style="margin: 1em 0;"></fast-divider>
      <div style="display: flex; flex-direction: row; justify-content: space-between; align-items: center;">
      <fast-flipper id="previous" onclick="${_previous}" direction="previous" disabled=${previous_disable}>
      </fast-flipper>
      <fast-flipper id="next" onclick="${_next}" disabled=${next_disable}>
      </fast-flipper>
      </div>
    </div>
    </div>
    """

    def __init__(self, **params):
        super().__init__(**params)
        for item in self.items:
            item.param.watch(self._ready, 'ready')
        self.items[0].active = True
        self._current = 0
        self.current = self.items[0]
        self.preview = pn.pane.JSON(self.spec, depth=-1, sizing_mode='stretch_both')
        self._modal_content = [
            '# Dashboard specification preview',
            FastDivider(),
            self.preview
        ]

    def _ready(self, event):
        self.next_disable = not event.obj.ready
        if event.obj.auto_advance:
            self._next()

    def open_modal(self):
        from .state import state
        self.preview.object = dict(state.spec)
        if state.modal.objects == [self.preview]:
            state.template.open_modal()
            return
        state.modal.loading = True
        state.template.open_modal()
        state.modal[:] = self._modal_content
        state.modal.loading = False

    def _previous(self, event=None):
        if self.previous_disable:
            return
        self._current -= 1
        self.current.active = False
        item = self.items[self._current]
        item.active = True
        self.current = item
        self.next_disable = False
        if self._current == 0:
            self.previous_disable = True

    def _next(self, event=None):
        if self.next_disable:
            return
        self.loading = True
        self._current += 1
        item = self.items[self._current]
        self.current.active = False
        item.active = True
        self.current = item
        self.previous_disable = False
        if self._current == (len(self.items)-1):
            self.next_disable = True
        self.loading = False



class WizardItem(ReactiveHTML):

    active = param.Boolean(default=False)

    auto_advance = param.Boolean(default=False)

    sizing_mode = param.String(default='stretch_width', readonly=True)

    ready = param.Boolean(default=False)

    spec = param.Dict(precedence=-1)

    __abstract = True

    def __init__(self, **params):
        super().__init__(**params)
        spec_params = [p for p in self.param if self.param[p].precedence == 1]
        if spec_params:
            self.param.watch(self._update_spec, spec_params)

    def _update_spec(self, *events):
        pass



import param

from lumen.config import _TEMPLATES

from .base import WizardItem


class ConfigEditor(WizardItem):

    title = param.String(doc="Enter a title")

    logo = param.String(doc="Path or URL pointing to a logo.")

    template = param.Selector(precedence=1, objects=list(_TEMPLATES),
                              doc="Select the template")

    theme = param.Selector(default='default', doc="Select the theme")

    layout = param.Selector(precedence=1)

    _template = """
      <span style="font-size: 2em"><b>Config Editor</b></span>
      <p>Declare the high-level configuration of the dashboard.</p>
      <fast-divider></fast-divider>
      <form style="display: grid; line-height: 2em;">
        <label for="title"><b>Title</b></label>
        <fast-text-field id="title" placeholder="{{ param.title.doc }}" value="${title}"></fast-text-field>
        <label for="Logo"><b>Logo</b></label>
        <fast-text-field id="logo" placeholder="{{ param.logo.doc }}" value="${logo}"></fast-text-field>
        <label for="template-${id}"><b>Template</b></label>
        <fast-select id="template" value="${template}">
          {% for tname in param.template.objects %}
          <fast-option value="{{ tname.lower() }}">{{ tname }}</fast-option>
          {% endfor %}
        </fast-select>
        <fast-tooltip anchor="template-${id}">{{ param.template.doc }}</fast-tooltip>
        <label for="theme-${id}"><b>Theme</b></label>
        <fast-select id="theme" value="${theme}">
          <fast-option value="default">Default</fast-option>
          <fast-option value="dark">Dark</fast-option>
        </fast-select>
        <fast-tooltip anchor="theme-${id}">{{ param.theme.doc }}</fast-tooltip>
      </form>
    """

    _dom_events = {'title': ['change']}

    @param.depends('spec', watch=True)
    def _update_params(self):
        update = {key: value for key, value in self.spec.items()
                  if key in self.param}
        self.param.set_param(**update)

    def _update_spec(self, *events):
        for event in events:
            if event.new:
                self.spec[event.name] = event.new



import os

from functools import partial

import panel as pn
import param

from panel.widgets import Widget as _PnWidget

from ..base import MultiTypeComponent
from ..state import state
from ..util import is_ref, resolve_module_reference

_PARAM_MAP = {
    dict : param.Dict,
    float: param.Number,
    list : param.List,
    str  : param.String
}


class Variables(param.Parameterized):
    """
    The Variables component stores a number Variable types and mirrors
    their values onto dynamically created parameters allowing other
    components to easily watch changes in a variable.
    """

    def __init__(self, **params):
        super().__init__(**params)
        self._vars = {}

    @classmethod
    def from_spec(cls, spec):
        variables = cls()
        if pn.state.curdoc:
            state._variables[pn.state.curdoc] = variables
        for name, var_spec in spec.items():
            if not isinstance(var_spec, dict):
                var_spec = {
                    'type': 'constant',
                    'default': var_spec
                }
            var = Variable.from_spec(dict(var_spec, name=name), variables)
            variables.add_variable(var)
        return variables

    def add_variable(self, var):
        """
        Adds a new variable to the Variables instance and sets up
        a parameter that can be watched.
        """
        self._vars[var.name] = var
        self.param.add_parameter(var.name, param.Parameter(default=var.value))
        var.param.watch(partial(self._update_value, var.name), 'value')

    def _update_value(self, name, event):
        self.param.update({name: event.new})

    def __getitem__(self, key):
        if key in self.param:
            return getattr(self, key)
        else:
            raise KeyError(f'No variable named {key!r} has been defined.')

    def panel(self, variables=None):
        if variables == []:
            return None
        column = pn.Column(name='Variables', sizing_mode='stretch_width')
        for key, var in self._vars.items():
            if variables is None or key in variables:
                var_panel = var.panel
                if var_panel is not None:
                    column.append(var_panel)
        return column


class Variable(MultiTypeComponent):
    """
    A Variable may declare a static or dynamic value that can be
    referenced from other components. The source of the Variable value
    can be anything from an environment variable to a widget or URL
    parameter. Variable components allow a concise way to configure
    other components and make it possible to orchestrate actions across
    multiple components.
    """

    default = param.Parameter(doc="""
       Default value to use if no other value is defined""")

    label = param.String(default=None, doc="""
       Optional label override for variable. Used wherever the variable
       shows up in the UI.""")

    materialize = param.Boolean(default=False, constant=True, doc="""
       Whether the variable should be inlined as a constant variable
       (in the Lumen Builder).""")

    required = param.Boolean(default=False, constant=True, doc="""
       Whether the variable should be treated as required.""")

    secure = param.Boolean(default=False, constant=True, doc="""
       Whether the variable should be treated as secure.""")

    value = param.Parameter(doc="""
       The materialized value of the variable.""")

    variable_type = None

    __abstract = True

    _valid_keys = 'params'
    _validate_params = True

    def __init__(self, **params):
        if 'value' not in params and 'default' in params:
            params['value'] = params['default']
        if 'label' not in params and 'name' in params:
            params['label'] = params['name']
        super().__init__(**params)
        self.param.watch(self._update_value_from_default, 'default')

    def _update_value_from_default(self, event):
        if event.old is self.value or self.value is self.param.value.default:
            self.value = event.new

    @classmethod
    def from_spec(cls, spec, variables=None):
        if isinstance(spec, dict):
            spec = dict(spec)
            var_type = spec.pop('type')
        else:
            var_type = 'constant'
            spec = {'default': spec}
        var_type = cls._get_type(var_type)
        resolved_spec, refs = {}, {}
        for k, val in spec.items():
            if is_ref(val):
                refs[k] = val
                val = state.resolve_reference(val, variables)
                if isinstance(val, Variable):
                    val = val.value
            resolved_spec[k] = val
        return var_type(refs=refs, **resolved_spec)

    def as_materialized(self):
        """
        If the variable is to be materialized by the builder this
        implements the conversion from a variable that references
        some external value to a materialized value.
        """
        return Constant(default=self.value)

    @property
    def panel(self):
        """
        Optionally returns a renderable component that allows
        controlling the variable.
        """
        return None


class Constant(Variable):
    """
    A constant value variable.
    """

    variable_type = 'constant'


class EnvVariable(Variable):
    """
    An environment variable.
    """

    key = param.String(default=None)

    variable_type = 'env'

    def __init__(self, **params):
        super().__init__(**params)
        key = self.key or self.name
        if key in os.environ:
            self.value = os.environ[key]


class Widget(Variable):
    """
    A Widget variable that updates when the widget value changes.
    """

    throttled = param.Boolean(default=True, doc="""
       If the widget has a value_throttled parameter use that instead,
       ensuring that no intermediate events are generated, e.g. when
       dragging a slider.""")

    variable_type = 'widget'

    def __init__(self, **params):
        default = params.pop('default', None)
        refs = params.pop('refs', {})
        throttled = params.pop('throttled', True)
        label = params.pop('label', None)
        super().__init__(
            default=default, refs=refs, name=params.get('name'), label=label, throttled=throttled
        )
        kind = params.pop('kind', None)
        if kind is None:
            raise ValueError("A Widget Variable type must declare the kind of widget.")
        if '.' in kind:
            widget_type = resolve_module_reference(kind, _PnWidget)
        else:
            widget_type = getattr(pn.widgets, kind)
        if 'value' not in params and default is not None:
            params['value'] = default
        if self.label:
            params['name'] = self.label
        deserialized = {}
        for k, v in params.items():
            if k in widget_type.param:
                try:
                    v = widget_type.param[k].deserialize(v)
                except Exception:
                    pass
            deserialized[k] = v
        self._widget = widget_type(**deserialized)
        if self.throttled and 'value_throttled' in self._widget.param:
            self._widget.link(self, value_throttled='value')
            self.param.watch(lambda e: self._widget.param.update({'value': e.new}), 'value')
        else:
            self._widget.link(self, value='value', bidirectional=True)

    @property
    def panel(self):
        return self._widget


class URLQuery(Variable):
    """
    A variable obtained from the URL query parameters.
    """

    key = param.String(default=None)

    variable_type = 'url'

    def __init__(self, **params):
        super().__init__(**params)
        key = self.key or self.name
        if pn.state.location:
            pn.state.location.param.watch(self._update_value, 'search')
            self._update_value()
        if pn.state.session_args and key in pn.state.session_args:
            self.value = pn.state.session_args[key][0].decode('utf-8')

    def _update_value(self, *args):
        key = self.key or self.name
        self.value = pn.state.location.query_params.get(key, self.default)


class Cookie(Variable):
    """
    A variable obtained from the cookies in the request.
    """

    key = param.String(default=None)

    variable_type = 'cookie'

    def __init__(self, **params):
        super().__init__(**params)
        if pn.state.cookies:
            key = self.key or self.name
            self.value = pn.state.cookies.get(key, self.default)


class UserInfo(Variable):
    """
    A variable obtained from OAuth provider's user info.
    """

    key = param.String(default=None)

    variable_type = 'user'

    def __init__(self, **params):
        super().__init__(**params)
        if pn.state.user_info:
            key = self.key or self.name
            self.value = pn.state.user_info.get(key, self.default)


class Header(Variable):
    """
    A variable obtained from the request header.
    """

    key = param.String(default=None)

    variable_type = 'header'

    def __init__(self, **params):
        super().__init__(**params)
        if pn.state.headers:
            key = self.key or self.name
            self.value = pn.state.headers.get(key, self.default)
